"""
Wrapper for AWS Lambda.
"""

from __future__ import absolute_import
import os
import traceback
import time
import copy
import functools
import warnings
import collections
from uuid import uuid4

import epsagon.trace
import epsagon.runners.aws_lambda
import epsagon.triggers.aws_lambda
import epsagon.wrappers.python_function
import epsagon.utils
from epsagon.constants import STEP_DICT_NAME, EPSAGON_EVENT_ID_KEY
import epsagon.runners.python_function
from epsagon.common import EpsagonWarning
from .. import constants


def _add_status_code(runner, return_value):
    """
    Tries to extract the status code from the return value and add it
    as a metadata field
    :param runner: Runner event to update
    :param return_value: The return value to extract from
    """
    if isinstance(return_value, collections.Mapping):
        status_code = return_value.get('statusCode')
        if status_code:
            runner.resource['metadata']['status_code'] = status_code


# pylint: disable=too-many-statements
def lambda_wrapper(func):
    """Epsagon's Lambda wrapper."""

    # avoid double instrumentation
    if getattr(func, '__instrumented__', False):
        return func

    @functools.wraps(func)
    def _lambda_wrapper(*args, **kwargs):
        """
        Generic Lambda function wrapper
        """
        cold_start_duration = time.time() - constants.COLD_START_TIME
        trace = epsagon.trace.trace_factory.get_or_create_trace()
        trace.prepare()

        try:
            event, context = args
        except ValueError:
            # This can happen when someone manually calls handler without
            # parameters / sends kwargs. In such case we ignore this trace.
            return func(*args, **kwargs)

        if os.environ.get(
            'AWS_LAMBDA_INITIALIZATION_TYPE'
        ) == 'provisioned-concurrency':
            constants.COLD_START = False

        try:
            runner = epsagon.runners.aws_lambda.LambdaRunner(
                time.time(),
                context
            )
            trace.set_runner(runner)
        # pylint: disable=W0703
        except Exception as exception:
            # Regress to python runner.
            warnings.warn(
                'Lambda context is invalid, using simple python wrapper',
                EpsagonWarning
            )
            trace.add_exception(
                exception,
                traceback.format_exc()
            )
            return epsagon.wrappers.python_function.wrap_python_function(
                func,
                args,
                kwargs
            )

        if constants.COLD_START:
            runner.resource['metadata'][
                'aws.lambda.cold_start_duration'
            ] = cold_start_duration
        constants.COLD_START = False

        try:
            trace.add_event(
                epsagon.triggers.aws_lambda.LambdaTriggerFactory.factory(
                    time.time(),
                    event,
                    context
                )
            )
        # pylint: disable=W0703
        except Exception as exception:
            trace.add_exception(
                exception,
                traceback.format_exc(),
                additional_data={'event': event}
            )

        if not trace.disable_timeout_send:
            trace.set_timeout_handler(context)

        result = None
        try:
            result = func(*args, **kwargs)
            if trace.propagate_lambda_id and isinstance(result, dict):
                result[EPSAGON_EVENT_ID_KEY] = runner.event_id
                runner.resource['metadata']['propagation_enabled'] = True
            return result
        # pylint: disable=W0703
        except Exception as exception:
            runner.set_exception(
                exception,
                traceback.format_exc(),
                handled=False
            )
            raise
        finally:
            try:
                _add_status_code(runner, result)
                if not trace.metadata_only:
                    runner.resource['metadata']['return_value'] = result
            # pylint: disable=W0703
            except Exception as exception:
                trace.add_exception(
                    exception,
                    traceback.format_exc(),
                )
            try:
                if not trace.disable_timeout_send:
                    epsagon.trace.Trace.reset_timeout_handler()
            # pylint: disable=W0703
            except Exception:
                pass
            try:
                epsagon.trace.trace_factory.send_traces()
            # pylint: disable=W0703
            except Exception:
                epsagon.utils.print_debug('Failed to send Lambda trace')

    _lambda_wrapper.__instrumented__ = True
    return _lambda_wrapper


# pylint: disable=too-many-statements
def step_lambda_wrapper(func):
    """Epsagon's Step Lambda wrapper."""

    @functools.wraps(func)
    def _lambda_wrapper(*args, **kwargs):
        """
        Generic Step Function wrapper
        """
        cold_start_duration = time.time() - constants.COLD_START_TIME
        trace = epsagon.trace.trace_factory.get_or_create_trace()
        trace.prepare()

        try:
            event, context = args
        except ValueError:
            # This can happen when someone manually calls handler without
            # parameters / sends kwargs. In such case we ignore this trace.
            return func(*args, **kwargs)

        try:
            runner = epsagon.runners.aws_lambda.StepLambdaRunner(
                time.time(),
                context
            )
            trace.set_runner(runner)
        # pylint: disable=W0703
        except Exception as exception:
            # Regress to python runner.
            warnings.warn(
                'Lambda context is invalid, using simple python wrapper',
                EpsagonWarning
            )
            trace.add_exception(
                exception,
                traceback.format_exc()
            )
            return epsagon.wrappers.python_function.wrap_python_function(
                func,
                args,
                kwargs
            )

        if constants.COLD_START:
            runner.resource['metadata'][
                'aws.lambda.cold_start_duration'
            ] = cold_start_duration
        constants.COLD_START = False

        try:
            trace.add_event(
                epsagon.triggers.aws_lambda.LambdaTriggerFactory.factory(
                    time.time(),
                    event,
                    context
                )
            )
        # pylint: disable=W0703
        except Exception as exception:
            trace.add_exception(
                exception,
                traceback.format_exc(),
                additional_data={'event': event}
            )

        trace.set_timeout_handler(context)

        result = None
        try:
            result = func(*args, **kwargs)
            steps_data = epsagon.utils.find_in_object(
                event,
                STEP_DICT_NAME
            )

            if isinstance(result, dict):
                epsagon.utils.print_debug(
                    'Step function result type is dict, steps_data={}'.format(
                        steps_data
                    )
                )
                # If the step functions data is not present, then this is the
                # First step.
                if steps_data is None:
                    epsagon.utils.print_debug(
                        'Could not find existing steps data'
                    )
                    steps_dict = {'id': str(uuid4()), 'step_num': 0}
                    path = []
                # Otherwise, just advance the steps number by one.
                else:
                    # don't change trigger data
                    steps_dict, path = steps_data
                    steps_dict = copy.deepcopy(steps_dict)
                    if 'step_num' in steps_dict:
                        steps_dict['step_num'] += 1
                        epsagon.utils.print_debug(
                            'Steps data found, new dict={}'.format(steps_dict)
                        )
                    else:
                        steps_dict = {'id': str(uuid4()), 'step_num': 0}
                        epsagon.utils.print_debug(
                            'Steps data not found, new dict={}'.format(
                                steps_dict
                            )
                        )

                result_path = result
                # Tries to inject the steps data in the configured
                # or same path where it was found
                if isinstance(trace.step_dict_output_path, list):
                    path = trace.step_dict_output_path
                try:
                    for sub_path in path:
                        result_path = result_path.get(sub_path)
                except Exception as exception:  # pylint: disable=broad-except
                    epsagon.utils.print_debug(
                        'Could not put steps in path={}'.format(path)
                    )
                if result_path:
                    epsagon.utils.print_debug(
                        'Adding steps dict to result_path={}'.format(
                            result_path
                        )
                    )
                    result_path[STEP_DICT_NAME] = steps_dict
                else:
                    epsagon.utils.print_debug(
                        'Adding steps dict to root result'
                    )
                    result[STEP_DICT_NAME] = steps_dict

                runner.add_step_data(steps_dict)
            return result
        # pylint: disable=W0703
        except Exception as exception:
            runner.set_exception(
                exception,
                traceback.format_exc(),
                handled=False
            )
            raise
        finally:
            try:
                _add_status_code(runner, result)
                if not trace.metadata_only:
                    runner.resource['metadata']['return_value'] = (
                        copy.deepcopy(result)
                    )
            # pylint: disable=W0703
            except Exception as exception:
                trace.add_exception(
                    exception,
                    traceback.format_exc(),
                )
            try:
                epsagon.trace.Trace.reset_timeout_handler()
            # pylint: disable=W0703
            except Exception:
                pass
            try:
                epsagon.trace.trace_factory.send_traces()
            # pylint: disable=W0703
            except Exception:
                pass

    return _lambda_wrapper



"""
Trace object holds events and metadata
"""
# pylint: disable=too-many-lines

from __future__ import absolute_import, print_function
import os
import sys
import time
import decimal
import traceback
import warnings
import signal
import threading
import random
import json
import urllib3.exceptions

from epsagon.event import BaseEvent
from epsagon.common import EpsagonWarning, ErrorCode
from epsagon.trace_encoder import TraceEncoder
from epsagon.trace_transports import NoneTransport, HTTPTransport, LogTransport
from .constants import (
    TIMEOUT_GRACE_TIME_MS,
    MAX_LABEL_SIZE,
    DEFAULT_SAMPLE_RATE,
    TRACE_URL_PREFIX,
    LAMBDA_TRACE_URL_PREFIX,
    is_strong_key,
    __version__
)

MAX_EVENTS_PER_TYPE = 20
MAX_TRACE_SIZE_BYTES = 64 * (2 ** 10)
DEFAULT_MAX_TRACE_SIZE_BYTES = 64 * (2 ** 10)
MAX_METADATA_FIELD_SIZE_LIMIT = 1024 * 3
FAILED_TO_SERIALIZE_MESSAGE = 'Failed to serialize returned object to JSON'


# pylint: disable=invalid-name
class _number_str(float):
    """ Taken from `bootstrap.py` of AWS Lambda Python runtime """
    # pylint: disable=super-init-not-called
    def __init__(self, o):
        self.o = o

    def __repr__(self):
        return str(self.o)


def _decimal_serializer(o):
    """ Taken from `bootstrap.py` of AWS Lambda Python runtime """
    if isinstance(o, decimal.Decimal):
        return _number_str(o)
    if isinstance(o, bytes):
        return o.decode('utf-8')
    raise TypeError(repr(o) + ' is not JSON serializable')


def get_thread_id():
    """
    Return current thread id
    :return: thread id
    """
    return threading.currentThread().ident


def create_transport(collector_url, token):
    if (os.getenv('EPSAGON_LOG_TRANSPORT') or '').upper() == 'TRUE':
        return LogTransport()
    return HTTPTransport(collector_url, token)


# pylint: disable=R0904
class TraceFactory(object):
    """
    A trace factory.
    """

    LOCK = threading.Lock()

    def __init__(self):
        """
        Initialize.
        """
        self.traces = {}
        self.app_name = ''
        self.token = ''
        self.collector_url = ''
        self.metadata_only = True
        self.disable_timeout_send = False
        self.debug = False
        self.send_trace_only_on_error = False
        self.url_patterns_to_ignore = None
        self.keys_to_ignore = None
        self.keys_to_allow = None
        self.use_single_trace = True
        self.singleton_trace = None
        self.local_thread_to_unique_id = {}
        self.transport = NoneTransport()
        self.split_on_send = False
        self.disabled = False
        self.propagate_lambda_id = False
        self.logging_tracing_enabled = False
        self.step_dict_output_path = None
        self.sample_rate = DEFAULT_SAMPLE_RATE

    def initialize(
        self,
        app_name,
        token,
        collector_url,
        metadata_only,
        disable_timeout_send,
        debug,
        send_trace_only_on_error,
        url_patterns_to_ignore,
        keys_to_ignore,
        keys_to_allow,
        transport,
        split_on_send,
        propagate_lambda_id,
        logging_tracing_enabled,
        step_dict_output_path,
        sample_rate,
    ):
        """
        Initializes The factory with user's data.
        User can configure here trace parameters.
        :param app_name: application name
        :param token: user's token
        :param collector_url: the url to send traces to.
        :param metadata_only: whether to send metadata only or not.
        :param disable_timeout_send: whether to disable traces send on timeout
         (when enabled, is done using a signal handler).
        :param debug: debug flag.
        :param send_trace_only_on_error: Whether to send trace only when
         there is error or not.
        :param url_patterns_to_ignore: URL patterns to ignore in HTTP data
         collection.
        :param keys_to_ignore: List of keys to ignore while extracting metadata.
        :param keys_to_allow: List of keys to allow while extracting metadata.
        :param split_on_send: Split trace into multiple traces in case it's size
         exceeds the maximum size.
        :param logging_tracing_enabled:
            Add an epsagon log id to all loggings and prints
        :param step_dict_output_path:
            Path in the result dict to append the Epsagon steps data
        :param sample_rate: A number between 0 and 1, represents the
            probability of a trace to be sent.
            When enabled (value < 1), sampling will be performed
            according to the given value.
        :return: None
        """
        self.app_name = app_name
        self.token = token
        self.collector_url = collector_url
        self.metadata_only = metadata_only
        self.disable_timeout_send = disable_timeout_send
        self.debug = debug
        self.send_trace_only_on_error = send_trace_only_on_error
        self.url_patterns_to_ignore = (
            set(url_patterns_to_ignore) if url_patterns_to_ignore else set()
        )
        self.keys_to_ignore = [] if keys_to_ignore is None else keys_to_ignore
        self.keys_to_allow = [] if keys_to_allow is None else keys_to_allow
        self.transport = transport
        self.split_on_send = split_on_send
        self.propagate_lambda_id = propagate_lambda_id
        self.logging_tracing_enabled = logging_tracing_enabled
        self.step_dict_output_path = step_dict_output_path
        self.sample_rate = sample_rate
        self.update_tracers()

    def update_tracers(self):
        """
        Update tracers to have latest parameters in case of re-initialization
        of the factory.
        """
        tracers_to_update = (
            [self.singleton_trace, ] if self.singleton_trace else
            self.traces.values()
        )

        for tracer in tracers_to_update:
            tracer.app_name = self.app_name
            tracer.token = self.token
            tracer.collector_url = self.collector_url
            tracer.metadata_only = self.metadata_only
            tracer.disable_timeout_send = self.disable_timeout_send
            tracer.debug = self.debug
            tracer.send_trace_only_on_error = self.send_trace_only_on_error
            tracer.url_patterns_to_ignore = self.url_patterns_to_ignore
            tracer.keys_to_ignore = self.keys_to_ignore
            tracer.keys_to_allow = self.keys_to_allow
            tracer.transport = self.transport
            tracer.split_on_send = self.split_on_send
            tracer.propagate_lambda_id = self.propagate_lambda_id
            tracer.logging_tracing_enabled = self.logging_tracing_enabled
            tracer.step_dict_output_path = self.step_dict_output_path
            tracer.sample_rate = self.sample_rate

    def switch_to_multiple_traces(self):
        """
        Set the use_single_trace flag to False.
        :return: None
        """
        self.use_single_trace = False

    def _create_new_trace(self, unique_id=None):
        """
        Creating new trace instance
        :param unique_id: trace unique id
        :return: new trace
        """
        return Trace(
            app_name=self.app_name,
            token=self.token,
            collector_url=self.collector_url,
            metadata_only=self.metadata_only,
            disable_timeout_send=self.disable_timeout_send,
            debug=self.debug,
            send_trace_only_on_error=self.send_trace_only_on_error,
            url_patterns_to_ignore=self.url_patterns_to_ignore,
            keys_to_ignore=self.keys_to_ignore,
            keys_to_allow=self.keys_to_allow,
            transport=self.transport,
            split_on_send=self.split_on_send,
            propagate_lambda_id=self.propagate_lambda_id,
            logging_tracing_enabled=self.logging_tracing_enabled,
            step_dict_output_path=self.step_dict_output_path,
            sample_rate=self.sample_rate,
            unique_id=unique_id,
        )

    def get_or_create_trace(self, unique_id=None):
        """
        Gets or create a trace - thread-safe
        :param unique_id: unique id
        :return: trace
        """
        return self._get_trace(unique_id=unique_id, should_create=True)

    def _get_thread_trace(self, should_create=False):
        """
        Get or create trace assuming multi-threaded tracer.
        :return: The trace.
        """
        # If multiple threads are used, then create a new trace for each thread
        thread_id = get_thread_id()
        if thread_id not in self.traces:
            if not should_create:
                return None
            new_trace = self._create_new_trace()
            self.traces[thread_id] = new_trace
        return self.traces[thread_id]

    def _get_trace(self, unique_id=None, should_create=False):
        """
        Get trace based on the tracing mode:
        single trace (use_single_trace=True) or
        multiple threads (use_single_trace=False)

        If should_create then creating a new trace if trace does not exist
        if use_single_trace is set to False, each thread will have
        it's own trace..
        :return: The trace.
        """
        with TraceFactory.LOCK:
            unique_id = self.get_thread_local_unique_id(unique_id)
            if unique_id:
                trace = (
                    self.singleton_trace
                    if self.singleton_trace and not self.traces
                    else self.traces.get(
                            unique_id, None
                    )
                )
                if not trace:
                    if not should_create:
                        return None
                    trace = self._create_new_trace(unique_id)
                # Making sure singleton trace contains the latest trace
                trace.unique_id = unique_id
                self.singleton_trace = trace
                self.traces[unique_id] = trace
                return trace

            if self.use_single_trace:
                if self.singleton_trace is None and should_create:
                    self.singleton_trace = self._create_new_trace()
                return self.singleton_trace

            # If multiple threads are used, then create a
            # new trace for each thread
            return self._get_thread_trace(should_create=should_create)

    @property
    def active_trace(self):
        """
        Return the active trace
        :return: None
        """
        local_unique_id = self.get_thread_local_unique_id()
        return self.traces.get(local_unique_id, self.singleton_trace)

    def switch_active_trace(self, unique_id):
        """
        Sets the active trace by unique id
        :return: unique id
        """
        with self.LOCK:
            trace = self.traces.get(unique_id, None)
            if trace:
                self.singleton_trace = trace
            return trace

    def pop_trace(self, trace=None):
        """
        Sets the active trace by unique id
        :return: unique id
        """
        with self.LOCK:
            if self.traces:
                trace = self.traces.pop(self.get_trace_identifier(trace), None)
                if not self.traces:
                    self.singleton_trace = None
                return trace
            if not trace:
                trace = self.singleton_trace
                self.singleton_trace = None
                return trace
            return None

    def get_thread_local_unique_id(self, unique_id=None):
        """
        Get thread local unique id
        :param unique_id: input unique id
        :return: active id if there's an active unique id or given one
        """
        return self.local_thread_to_unique_id.get(
            get_thread_id(), unique_id
        )

    def set_thread_local_unique_id(self, unique_id=None):
        """
        Set thread local unique id
        :param unique_id: input unique id
        :return: the active unique id
        """
        unique_id = (
            unique_id if unique_id else (
                self.singleton_trace.unique_id if self.singleton_trace else None
            )
        )
        self.local_thread_to_unique_id[get_thread_id()] = unique_id
        return unique_id

    def unset_thread_local_unique_id(self):
        """
        Unset thread local unique id
        :return: None
        """
        self.local_thread_to_unique_id.pop(get_thread_id(), None)

    def get_trace_identifier(self, trace=None):
        """
        Return the trace identifier
        :return: trace identifier
        """
        if not trace:
            trace = self.singleton_trace

        return (
            get_thread_id()
            if not trace or not trace.unique_id
            else trace.unique_id
        )

    def get_trace(self):
        """
        Get the relevant trace (may be thread-based or a singleton trace)
        :return: The trace, None if trace does not exist
        """
        return self._get_trace(should_create=False)

    def add_event(self, event):
        """
        Add  event to the relevant trace.
        :param event: The event to add.
        :return: None
        """
        trace = self.get_trace()
        if trace:
            trace.add_event(event)

    def set_runner(self, event):
        """
        Add a runner event to the relevant trace.
        :param event: The event to add.
        :return: None
        """
        if self.get_trace():
            self.get_trace().set_runner(event)

    def add_exception(self, exception, stack_trace, additional_data=''):
        """
        add an exception to the relevant trace.
        :param exception: the exception to add
        :param stack_trace: the traceback at the moment of the event
        :param additional_data: a json serializable object that contains
            additional data regarding the exception
        :return: None
        """
        if self.debug:
            print('[EPSAGON_DEBUG] Epsagon exception: {}\n{}-----'.format(
                exception,
                stack_trace
            ))
        if self.get_trace():
            self.get_trace().add_exception(
                exception,
                stack_trace,
                additional_data
            )

    def add_label(self, key, value):
        """
        Add label to the current thread's trace.
        :param key:
        :param value:
        """
        if self.get_trace():
            self.get_trace().add_label(key, value)

    def is_logging_tracing_enabled(self):
        """
        Get the value of the logging_tracing_enabled flag
        """
        return self.logging_tracing_enabled

    def get_log_id(self):
        """
        Get the log id of the current trace
        """
        if self.get_trace():
            return self.get_trace().get_log_id()
        return None

    def set_error(self, exception, traceback_data=None, from_logs=False):
        """
        Set an error for the current thread's trace.
        :param exception: The exception
        :param traceback_data: The traceback data.
        :param from_logs: True if the exception was captured from logging
        """
        if self.get_trace():
            self.get_trace().set_error(
                exception,
                traceback_data,
                from_logs=from_logs
            )

    def get_trace_url(self):
        """
        Return the trace URL based on the runner ID.
        :return: string of the trace URL if exists, otherwise None
        """
        url = None
        if self.get_trace():
            url = self.get_trace().get_trace_url()
        return url

    def send_traces(self, trace=None):
        """
        Send the traces for the current thread.
        :return: None
        """
        if self.disabled:
            print('EPSAGON: Trace not sent (disabled).')
            return

        trace = trace if trace else self.get_trace()

        if trace:
            trace_sent = False
            try:
                trace.send_traces()
                trace_sent = True
                self.pop_trace(trace=trace)
            except Exception:  # pylint: disable=W0703
                if not trace_sent:
                    self.pop_trace(trace=trace)

    def prepare(self):
        """
        Prepare the relevant trace.
        :return: None
        """
        trace = self.get_trace()
        if trace:
            trace.prepare()

    def enable(self):
        """
        Enables Epsagon
        :return: None
        """
        self.disabled = False

    def disable(self):
        """
        Disables Epsagon
        :return: None
        """
        self.disabled = True


# pylint: disable=too-many-public-methods
class Trace(object):
    """
    Represents runtime trace
    """

    def __init__(
        self,
        app_name='',
        token='',
        collector_url='',
        metadata_only=True,
        disable_timeout_send=False,
        debug=False,
        send_trace_only_on_error=False,
        url_patterns_to_ignore=None,
        keys_to_ignore=None,
        keys_to_allow=None,
        unique_id=None,
        split_on_send=False,
        transport=NoneTransport(),
        propagate_lambda_id=False,
        logging_tracing_enabled=False,
        step_dict_output_path=None,
        sample_rate=DEFAULT_SAMPLE_RATE,
    ):
        """
        initialize.
        """
        self.app_name = app_name
        self.unique_id = unique_id
        self.token = token
        self.events = []
        self.exceptions = []
        self.custom_labels = {}
        self.custom_labels_size = 0
        self.has_custom_error = False
        self.version = __version__
        self.collector_url = collector_url
        self.metadata_only = metadata_only
        self.disable_timeout_send = disable_timeout_send
        self.debug = debug
        self.send_trace_only_on_error = send_trace_only_on_error
        self.url_patterns_to_ignore = url_patterns_to_ignore
        self.transport = transport
        self.split_on_send = split_on_send
        self.propagate_lambda_id = propagate_lambda_id
        self.logging_tracing_enabled = logging_tracing_enabled
        self.step_dict_output_path = step_dict_output_path
        self.sample_rate = sample_rate

        if keys_to_ignore:
            self.keys_to_ignore = [self._strip_key(key) for key in
                                   keys_to_ignore]
            if self.debug:
                print(
                    'Setting keys_to_ignore={}'.format(keys_to_ignore)
                )
        else:
            self.keys_to_ignore = []
        if keys_to_allow:
            self.keys_to_allow = [self._strip_key(key) for key in keys_to_allow]
            if self.debug:
                print(
                    'Setting keys_to_allow={}'.format(keys_to_allow)
                )
        else:
            self.keys_to_allow = []
        self.platform = 'Python {}.{}'.format(
            sys.version_info.major,
            sys.version_info.minor
        )
        self.runner = None
        self.trace_sent = False

    # pylint: disable=unused-argument, unused-variable
    def timeout_handler(self, signum, frame):
        """
        Send a trace in case of timeout.
        Invoked by a pre-set alarm.
        """
        try:
            if self.debug:
                print('Epsagon timeout handler called. Stack trace:')
                traceback.print_stack(limit=100)

            self.runner.set_timeout()
            self.send_traces()

            # pylint: disable=W0703
        except Exception:
            pass

    @staticmethod
    def reset_timeout_handler():
        """
        Cancel an already set alarm.
        """
        signal.alarm(0)
        signal.signal(signal.SIGALRM, signal.SIG_DFL)

    def set_timeout_handler(self, context):
        """
        Sets a timeout handler for the current tracer.
        :param context: context, as received by Lambda.
        """
        try:
            if not hasattr(context, 'get_remaining_time_in_millis'):
                return

            original_timeout = context.get_remaining_time_in_millis()
            if original_timeout <= TIMEOUT_GRACE_TIME_MS:
                return

            modified_timeout = (
                           original_timeout - TIMEOUT_GRACE_TIME_MS) / 1000.0
            signal.setitimer(signal.ITIMER_REAL, modified_timeout)
            original_handler = signal.signal(
                signal.SIGALRM,
                self.timeout_handler
            )

            if (
                    original_handler and
                    original_handler is not self.timeout_handler
            ):
                warnings.warn(
                    'Epsagon Warning: Overriding existing '
                    'SIGALRM handler {!r}'.format(original_handler)
                )

        # pylint: disable=W0703
        except Exception:
            pass

    def add_exception(self, exception, stack_trace, additional_data=''):
        """
        add an exception to the trace
        :param exception: the exception to add
        :param stack_trace: the traceback at the moment of the event
        :param additional_data: a json serializable object that contains
            additional data regarding the exception
        :return: None
        """

        try:
            exception_dict = {
                'type': str(type(exception)),
                'message': str(exception),
                'traceback': stack_trace,
                'time': time.time(),
                'additional_data': additional_data
            }

            self.exceptions.append(exception_dict)
        # Making sure that tracing inner exception won't crash
        # pylint: disable=W0703
        except Exception:
            pass

    def prepare(self):
        """
        Prepares new trace.
        Prints error if token is empty, and empty events list.
        :return: None
        """

        if self.token == '':
            warnings.warn(
                'Epsagon Error: Please initialize token, data won\'t be sent.',
                EpsagonWarning
            )

        self.events = []
        self.exceptions = []
        self.custom_labels = {}
        self.custom_labels_size = 0
        self.has_custom_error = False
        self.runner = None
        self.trace_sent = False

    def initialize(
            self,
            app_name,
            token,
            collector_url,
            metadata_only,
            disable_timeout_send,
            debug
    ):
        """
        Initializes trace with user's data.
        User can configure here trace parameters.
        :param app_name: application name
        :param token: user's token
        :param collector_url: the url to send traces to.
        :param metadata_only: whether to send metadata only or not.
        :param disable_timeout_send: whether to disable traces send on timeout
         (when enabled, is done using a signal handler).
        :param debug: debug flag
        :return: None
        """

        self.app_name = app_name
        self.token = token
        self.collector_url = collector_url
        self.metadata_only = metadata_only
        self.disable_timeout_send = disable_timeout_send
        self.debug = debug

    @staticmethod
    def load_from_dict(trace_data):
        """
        Load new trace object from dict.
        :param trace_data: dict data of trace
        :return: Trace
        """
        trace = Trace()
        trace.app_name = trace_data['app_name']
        trace.token = trace_data['token']
        trace.version = trace_data['version']
        trace.platform = trace_data['platform']
        trace.exceptions = trace_data.get('exceptions', [])
        trace.events = []
        for event in trace_data['events']:
            trace.add_event(BaseEvent.load_from_dict(event))
        return trace

    def set_runner(self, runner):
        """
        Sets the runner of the current tracer
        :param runner: Runner to set
        """
        if self.logging_tracing_enabled:
            runner.resource['metadata']['logging_tracing_enabled'] = True

        self.add_event(runner, should_terminate=False)
        self.runner = runner

    def clear_events(self):
        """
        Clears the events list
        :return: None
        """
        self.events = []

    def add_event(self, event, should_terminate=True):
        """
        Add event to events list.
        :param event: BaseEvent
        :param should_terminate: If True, `event.terminate()` is called
        :return: None
        """
        if should_terminate:
            event.terminate()
        self.events.append(event)

    def verify_custom_label(self, key, value):
        """
        Verifies custom label is valid, both in size and type.
        :param key: Key for the label data (string)
        :param value: Value for the label data (string)
        :return: True/False
        """
        if not isinstance(key, str):
            print('EPSAGON: label key support only string type')
            print('Received {key}, key type={type}'.format(
                key=key,
                type=type(key)
            ))
            return False
        if not isinstance(value, (int, float, str)):
            print('EPSAGON: label value support only string, int, float types')
            print('Received {key}, value type={type}'.format(
                key=key,
                type=type(value)
            ))
            return False

        # Even for numeric types we are checking the length of the
        # stringified value.
        if (
                len(key) +
                len(str(value)) +
                self.custom_labels_size > MAX_LABEL_SIZE
        ):
            return False

        self.custom_labels_size += len(key) + len(str(value))

        return True

    def add_label(self, key, value):
        """
        Adds a custom label given by the user to the runner
        of the current trace
        :param key: Key for the label data (string)
        :param value: Value for the label data (string, bool, int, float)
        """
        if isinstance(value, dict):
            for dict_key, dict_value in value.items():
                self.add_label('{}.{}'.format(key, dict_key), dict_value)
            return

        if not self.verify_custom_label(key, value):
            return
        self.custom_labels[key] = value

    def get_log_id(self):
        """
        Get the log id if the logging_tracing_enabled flag is on,
        else return None
        """
        if self.logging_tracing_enabled and self.runner:
            trace_id = self.runner.resource['metadata'].get('trace_id')
            if trace_id:
                return 'E#' + str(trace_id) + '#E'

        return None

    def set_error(self, exception, traceback_data=None, from_logs=False):
        """
        Sets the error value of the runner
        :param exception: Exception object or String to set.
        :param traceback_data: traceback string
        :param from_logs: True if the exception was captured from logging
        """
        if not self.runner:
            return

        if not traceback_data:
            if getattr(exception, '__traceback__', None):
                traceback_data = ''.join(traceback.format_exception(
                    type(exception),
                    exception,
                    getattr(exception, '__traceback__'),
                ))
            else:
                traceback_data = ''.join(
                    traceback.format_list(traceback.extract_stack())
                )
        # Convert exception string to Exception type
        if isinstance(exception, str):
            exception = Exception(exception)

        self.runner.set_exception(
            exception,
            traceback_data,
            from_logs=from_logs
        )

    def update_runner_with_labels(self):
        """
        Adds the custom labels to the runner of the trace
        """
        if not self.custom_labels:
            return

        self.runner.resource['metadata']['labels'] = json.dumps(
            self.custom_labels
        )

    def to_dict(self):
        """
        Convert trace to dict.
        :return: Trace dict
        """
        try:
            self.update_runner_with_labels()
        # pylint: disable=W0703
        except Exception as exception:
            # Ignore custom logs in case of error.
            self.add_exception(
                exception,
                traceback.format_exc()
            )

        return {
            'token': self.token,
            'app_name': self.app_name,
            'events': [event.to_dict() for event in self.events],
            'exceptions': self.exceptions,
            'version': self.version,
            'platform': self.platform,
        }

    @staticmethod
    def trim_metadata(metadata):
        """
        Trimming metadata
        :param metadata: metadata
        :return:
        """
        for key in list(metadata.keys()):
            if not is_strong_key(key):
                metadata.pop(key)

    @staticmethod
    def _trim_dict_values(data, max_size):
        """
        Trimms dict values from a given data dict whose size bigger than
        the given max_size parameter
        If data is not a dict - the function does nothing.
        :param data: the data dict to trim fields from
        :param max_size: the max field size
        """
        if not isinstance(data, dict):
            return

        for field_name in data:
            value = data[field_name]
            if isinstance(value, dict):
                try:
                    json_value = json.dumps(value, default=_decimal_serializer)
                    if len(json_value) > max_size:
                        data[field_name] = json_value[:max_size]
                # pylint: disable=W0703
                except Exception:
                    data[field_name] = FAILED_TO_SERIALIZE_MESSAGE

    @staticmethod
    def events_sorter(event):
        """
        Events sort function
        :param event: event
        :return: sorting result
        """
        return 1 if event.origin in ['runner', 'trigger'] else 0

    @property
    def _max_trace_size(self):
        """
        Retreive the max trace size
        """
        max_trace_size = os.getenv('EPSAGON_MAX_TRACE_SIZE')
        if max_trace_size:
            try:
                return int(max_trace_size)
            except ValueError:
                print('Invalid max Epsagon trace size given')

        return DEFAULT_MAX_TRACE_SIZE_BYTES

    @property
    def length(self):
        json_trace = json.dumps(
            self.to_dict(),
            cls=TraceEncoder,
            ensure_ascii=True
        )
        return len(json_trace)

    def _strip(self, trace_length):
        """
        Strips a given trace from all operations
        """
        for event in sorted(self.events, key=Trace.events_sorter):
            event_metadata_length = (
                len(json.dumps(
                    event.resource.get('metadata', {}),
                    cls=TraceEncoder,
                    ensure_ascii=True
                ))
            )
            Trace.trim_metadata(event.resource['metadata'])
            trace_length -= event_metadata_length
            if trace_length < self._max_trace_size:
                break

    @staticmethod
    def _strip_key(key):
        """
        Strip a given key from spaces, dashes, and underscores.
        :param key: The key to strip.
        :return: Stripped key.
        """
        return (
            str(key).lower().replace('-', '').replace('_', '').replace(' ', '')
        )

    def remove_ignored_keys(self, input_dict):
        """
        Remove ignored keys recursively in input_dict.
        If an ignored key has been found in a dict, then the
        dict is copied (shallow copy) and the ignored key is removed.
        :param input_dict: Input dict to remove ignored keys from.
        :return: a dict without the the ignored keys
        """
        # pylint: disable=too-many-nested-blocks
        if not self.keys_to_ignore:
            return input_dict
        copied_dict = input_dict.copy()
        # Python 2 returns a list, while Python3 returns an iterator.
        for key in input_dict:
            if self._strip_key(key) in self.keys_to_ignore:
                copied_dict.pop(key)
                if self.debug:
                    print(
                        'Removed ignored key {}'.format(key)
                    )
            else:
                value = input_dict[key]
                if isinstance(value, dict):
                    copied_dict[key] = self.remove_ignored_keys(value)
                else:
                    if isinstance(value, str):
                        try:
                            json_value = json.loads(value)
                            if isinstance(json_value, dict):
                                copied_dict[key] = (json.dumps(
                                    self.remove_ignored_keys(json_value)
                                ))
                        except (TypeError, ValueError):
                            pass
        return copied_dict

    def get_dict_with_allow_keys(self, input_dict):
        """
        Keeping only the branches which contains allowed keys.
        Any branch that doesn't contain allowed key will be deleted.
        :param input_dict: Input dict to remove branches without allowed keys.
        :return: a dict that contains branches which contains allowed keys
        """
        copied_dict = input_dict.copy()
        for key in input_dict:
            is_key_in_whitelist = self._strip_key(key) in self.keys_to_allow
            if not is_key_in_whitelist:
                value = input_dict[key]
                if isinstance(value, dict):
                    child_dict = self.get_dict_with_allow_keys(value)
                    if not child_dict:
                        copied_dict.pop(key)
                    else:
                        copied_dict[key] = child_dict
                else:
                    copied_dict.pop(key)
        return copied_dict

    def send_traces(self):
        """
        Should NOT be called by any wrapper! this method should ONLY be called
        by TraceFactory.
        If trace size exceeds the maximum size, and split flag is on
        then split the trace into multiple traces.
        :return: None
        """
        if (
                self.split_on_send
                and len(self.events) > 1
                and self.length > self._max_trace_size
        ):
            self._send_trace_split()
        else:
            self._send_traces()

    def _send_trace_split(self):
        """
        Split trace into multiple traces and send them one after the other.
        This is done by manipulating the trace object while keeping the same
        runner.
        :param trace: the trace to send.
        """
        # Get only events (without runner)
        all_events = self.events[:]
        if self.runner:
            all_events.remove(self.runner)

        self.runner.resource['metadata']['fragment_seq'] = 1
        self.clear_events()
        self.add_event(self.runner)
        for event in all_events:
            self.add_event(event)
            if self.length > self._max_trace_size:
                self.events.pop()
                self._send_traces()
                self.runner.resource['metadata']['fragment_seq'] += 1
                self.trace_sent = False
                self.clear_events()
                self.add_event(self.runner)
                self.add_event(event)

        # If there are events to send (except for runner)
        if len(self.events) > 1:
            self._send_traces()

    # pylint: disable=W0703
    def _send_traces(self):
        """
        Send trace to collector.
        :return: None
        """
        if self.token == '' or self.trace_sent:
            return

        rand_num = random.uniform(0, 1)
        if (
                (self.send_trace_only_on_error or self.sample_rate < rand_num)
                and self.runner
                and self.runner.error_code == ErrorCode.OK
        ):
            if self.debug:
                print('Trace was omitted. sample rate is: {},'
                      'random value: {}'.format(self.sample_rate, rand_num))
            return

        trace = ''
        self.transport = (
            self.transport
            if not isinstance(self.transport, NoneTransport)
            else create_transport(self.collector_url, self.token)
        )

        try:
            # Update events resource metadata.
            for event in self.events:
                # Remove ignored keys.
                event.resource['metadata'] = self.remove_ignored_keys(
                    event.resource['metadata'])
                # Keep allowed keys.
                if self.keys_to_allow:
                    event.resource['metadata'] = self.get_dict_with_allow_keys(
                        event.resource['metadata'])
                type(self)._trim_dict_values(
                    event.resource['metadata'],
                    MAX_METADATA_FIELD_SIZE_LIMIT
                )

        except Exception as exception:
            if self.debug:
                print(
                    'Failed to send trace: '
                    'updating events resource metadata failed: {}'.format(
                        exception
                    )
                )
                traceback.print_exc()
            return

        try:
            if self.runner:
                self.runner.terminate()

            trace = json.dumps(
                self.to_dict(),
                cls=TraceEncoder,
                ensure_ascii=True
            )

            trace_length = len(trace)
            if trace_length > self._max_trace_size:
                # Trace too big.
                self._strip(trace_length)
                self.runner.resource['metadata']['is_trimmed'] = True

                trace = json.dumps(
                    self.to_dict(),
                    cls=TraceEncoder,
                    ensure_ascii=True
                )

            self.transport.send(self)
            self.trace_sent = True

            if self.debug:
                print('Trace sent (size: {})'.format(
                    len(trace)
                ))
        except (
            urllib3.exceptions.TimeoutError,
            urllib3.exceptions.MaxRetryError
        ):
            print('Failed to send trace (size: {}) (timeout)'.format(
                len(trace)
            ))
        except Exception as exception:
            print('Failed to send trace (size: {}): {}'.format(
                len(trace),
                exception
            ))
        finally:
            if self.debug:
                print('collector url:', self.collector_url)
                print('trace:', trace)

    def get_trace_url(self):
        """
        Return the trace URL based on the runner ID.
        Handling special case for Lambda functions.
        :return: string of the trace URL.
        """
        if not self.runner:
            warnings.warn(
                'Epsagon Warning: Could not share'
                'trace URL since runner is empty.'
            )
            return ''

        trace_url = TRACE_URL_PREFIX.format(
            id=self.runner.resource['metadata'].get('trace_id', ''),
            start_time=int(self.runner.start_time)
        )
        if self.runner.resource['type'] == 'lambda':
            runner_resource = self.runner.resource
            trace_url = LAMBDA_TRACE_URL_PREFIX.format(
                aws_account=runner_resource['metadata']['aws_account'],
                region=runner_resource['metadata']['region'],
                function_name=self.runner.resource['name'],
                request_id=self.runner.event_id,
                request_time=int(self.runner.start_time)
            )
        return trace_url

# pylint: disable=C0103
trace_factory = TraceFactory()



import mock
import json
import itertools
import epsagon.constants


@mock.patch(
    'time.time',
    side_effect=itertools.count(start=1)
)
def test_function_wrapper_sanity(_, trace_transport):
    retval = 'success'

    @epsagon.measure
    def measured_function():
        return retval

    @epsagon.python_wrapper(name='test-func')
    def wrapped_function():
        measured_function()
        return retval

    assert wrapped_function() == retval
    labels = json.loads(trace_transport.last_trace.events[0].resource['metadata']['labels'])
    assert labels['measured_function_duration'] == 1



"""
Cloud Object Storage events module.
"""

from __future__ import absolute_import
from uuid import uuid4
import traceback

from ..event import BaseEvent
from ..trace import trace_factory


class COSEvent(BaseEvent):
    """
    Represents base Cloud Object Storage event.
    """
    ORIGIN = 'tencent-cos'
    RESOURCE_TYPE = 'cos'

    # pylint: disable=W0613
    def __init__(self, wrapped, instance, args, kwargs, start_time, response,
                 exception):
        """
        Initialize the Cloud Object Storage event
        :param wrapped: wrapt's wrapped
        :param instance: wrapt's instance
        :param args: wrapt's args
        :param kwargs: wrapt's kwargs
        :param start_time: Start timestamp (epoch)
        :param response: response data
        :param exception: Exception (if happened)
        """

        super(COSEvent, self).__init__(start_time)
        self.event_id = 'cos-{}'.format(str(uuid4()))
        self.resource['name'] = kwargs['bucket']
        self.resource['operation'] = kwargs['method']
        self.resource['metadata'] = {
            # pylint: disable=protected-access
            'tencent.region': instance._conf._region,
            'tencent.cos.object_key': kwargs['url'].split('myqcloud.com/')[-1],
        }
        if response:
            self.resource['metadata'].update({
                'tencent.cos.request_id': response.headers['x-cos-request-id'],
                'tencent.status_code': response.status_code,
            })

        if exception is not None:
            self.resource['metadata'].update({
                'tencent.cos.request_id': exception.get_request_id(),
                'tencent.status_code': exception.get_status_code(),
            })
            self.set_exception(exception, traceback.format_exc())


class COSEventFactory(object):
    """
    Factory class, generates Cloud Object Storage event.
    """
    @staticmethod
    def create_event(wrapped, instance, args, kwargs, start_time, response,
                     exception):
        """
        Create a Cloud Object Storage event.
        """
        trace_factory.add_event(COSEvent(
            wrapped,
            instance,
            args,
            kwargs,
            start_time,
            response,
            exception
        ))



"""
PynamoDB patcher module.
"""

from __future__ import absolute_import
import wrapt
from epsagon.modules.general_wrapper import wrapper
from ..events.pynamodb import PynamoDBEventAdapter, PynamoDBVendoredEventAdapter


def _vendored_wrapper(wrapped, instance, args, kwargs):
    """
    General wrapper for PynamoDB instrumentation.
    :param wrapped: wrapt's wrapped
    :param instance: wrapt's instance
    :param args: wrapt's args
    :param kwargs: wrapt's kwargs
    :return: None
    """

    # Skip non DynamoDB requests
    if 'https://dynamodb.' not in args[0].url:
        return wrapped(*args, **kwargs)

    return wrapper(
        PynamoDBVendoredEventAdapter,
        wrapped,
        instance,
        args,
        kwargs
    )


def _wrapper(wrapped, instance, args, kwargs):
    """
    General wrapper for PynamoDB instrumentation.
    :param wrapped: wrapt's wrapped
    :param instance: wrapt's instance
    :param args: wrapt's args
    :param kwargs: wrapt's kwargs
    :return: None
    """
    return wrapper(PynamoDBEventAdapter, wrapped, instance, args, kwargs)


def patch():
    """
    Patch module.
    :return: None
    """
    try:
        wrapt.wrap_function_wrapper(
            'pynamodb.connection.base',
            'Connection._make_api_call',
            _wrapper
        )
    except AttributeError:
        pass

    try:
        wrapt.wrap_function_wrapper(
            'botocore.vendored.requests.sessions',
            'Session.send',
            _vendored_wrapper
        )
    except AttributeError:
        pass



"""
Automatically imports all available modules for patch
"""

from __future__ import absolute_import
import os
import sys
from importlib import import_module

MODULES = {}
IGNORE_MODULES = ('__init__',)
PYTHON_EXTENSIONS = ('.py', '.pyc')
VERSION_DEPENDENCIES = {
    'fastapi': (3, 5, 3),
}

for module_name in os.listdir(os.path.dirname(__file__)):
    filename, ext = os.path.splitext(module_name)
    if filename in IGNORE_MODULES or ext not in PYTHON_EXTENSIONS:
        continue

    # Verify that the loaded module meets the minimum Python version
    if filename in VERSION_DEPENDENCIES:
        if sys.version_info < VERSION_DEPENDENCIES[filename]:
            continue

    try:
        imported = import_module('.{}'.format(filename), __name__)
        MODULES[filename] = imported
    except ImportError:
        pass



"""
sqlalchemy events module.
"""

from __future__ import absolute_import
import traceback
from uuid import uuid4

from ..trace import trace_factory
from ..event import BaseEvent
from ..utils import database_connection_type


class SqlAlchemyEvent(BaseEvent):
    """
    Represents base SqlAlchemy event.
    """

    ORIGIN = 'sqlalchemy'
    RESOURCE_TYPE = 'database'

    # pylint: disable=W0613
    def __init__(self, wrapped, instance, args, kwargs, start_time, response,
                 exception, operation):
        """
        Initialize.
        :param wrapped: wrapt's wrapped
        :param instance: wrapt's instance
        :param args: wrapt's args
        :param kwargs: wrapt's kwargs
        :param start_time: Start timestamp (epoch)
        :param response: response data
        :param exception: Exception (if happened)
        :param operation: sqlalchemy operation
        """
        super(SqlAlchemyEvent, self).__init__(start_time)

        self.event_id = 'sqlalchemy-{}'.format(str(uuid4()))

        self.resource['name'] = (
                instance.bind.url.database or instance.bind.url.host
        )
        self.resource['operation'] = operation

        # override event type with the specific DB type
        self.resource['type'] = database_connection_type(
            instance.bind.url.host,
            self.RESOURCE_TYPE
        )

        if exception is not None:
            self.set_exception(exception, traceback.format_exc())


class SqlAlchemyEventFactory(object):
    """
    Factory class, generates sqlalchemy event.
    """

    OPERATION_MAPPING = {
        '__init__': 'initialize',
        'close': 'close',
    }

    @staticmethod
    def create_event(wrapped, instance, args, kwargs, start_time, response,
                     exception):
        """
        Create sqlalchemy initialize/close event.
        """
        operation = SqlAlchemyEventFactory.OPERATION_MAPPING.get(
            getattr(wrapped, '__name__')
        )

        if not operation:
            return

        event = SqlAlchemyEvent(
            wrapped,
            instance,
            args,
            kwargs,
            start_time,
            response,
            exception,
            operation
        )

        trace_factory.add_event(event)



"""
Epsagon's init.
"""

from __future__ import absolute_import
import os
from .utils import init, print_debug
from .patcher import patch_all
from .constants import __version__, EPSAGON_HANDLER
from .trace import trace_factory
from .wrappers.custom import measure

if os.getenv(EPSAGON_HANDLER):
    from .handler import wrapper


def auto_load(_):
    """
    Called when setting `AUTOWRAPT_BOOTSTRAP=epsagon` (auto-tracing).
    """
    print_debug('Initialized using auto-tracing')
    init()


def dummy_wrapper(func):
    """
    A dummy wrapper for when Epsagon is disabled
    :param func: The function to wrap
    :return: The same function, unchanged
    """
    return func


def dummy_python_wrapper(*args, **_kwargs):
    """
    A dummy wrapper for when Epsagon is disabled.
    Used for general python functions
    :return: The same function, unchanged
    """
    def _inner_wrapper(func):
        return func

    if len(args) == 1 and callable(args[0]):
        return _inner_wrapper(args[0])

    return _inner_wrapper


if (
        os.getenv('DISABLE_EPSAGON') and
        os.getenv('DISABLE_EPSAGON').upper() == 'TRUE'
):
    os.environ['DISABLE_EPSAGON_PATCH'] = 'TRUE'
    lambda_wrapper = dummy_wrapper  # pylint: disable=C0103
    step_lambda_wrapper = dummy_wrapper  # pylint: disable=C0103
    chalice_wrapper = dummy_wrapper  # pylint: disable=C0103
    azure_wrapper = dummy_wrapper  # pylint: disable=C0103
    python_wrapper = dummy_python_wrapper  # pylint: disable=C0103
    flask_wrapper = dummy_wrapper  # pylint: disable=C0103
    gcp_wrapper = dummy_wrapper  # pylint: disable=C0103
    tencent_function_wrapper = dummy_wrapper  # pylint: disable=C0103
else:
    # Environments.
    from .wrappers import (
        lambda_wrapper,
        step_lambda_wrapper,
        chalice_wrapper,
        azure_wrapper,
        python_wrapper,
        gcp_wrapper,
        tencent_function_wrapper
    )

    # Frameworks.
    try:
        from .wrappers.flask import FlaskWrapper as flask_wrapper
    except ImportError:
        flask_wrapper = dummy_wrapper


# pylint: disable=C0103
label = trace_factory.add_label
error = trace_factory.set_error
disable = trace_factory.disable
enable = trace_factory.enable
get_trace_url = trace_factory.get_trace_url


__all__ = [
    'lambda_wrapper',
    'azure_wrapper',
    'python_wrapper',
    'init',
    'step_lambda_wrapper',
    'flask_wrapper',
    'wrapper',
    'gcp_wrapper',
    'tencent_function_wrapper',
    'chalice_wrapper',
    'auto_load',
    'measure',
]


# The modules are patched only if DISABLE_EPSAGON_PATCH variable is NOT 'TRUE'
if (
        not os.getenv('DISABLE_EPSAGON_PATCH') or
        os.getenv('DISABLE_EPSAGON_PATCH').upper() != 'TRUE'
):
    patch_all()



"""
Wrapper for Tencent Serverless Cloud Functions.
"""

from __future__ import absolute_import
import traceback
import time
import functools
import warnings
import collections

import epsagon.trace
import epsagon.runners.tencent_function
import epsagon.wrappers.python_function
import epsagon.utils
import epsagon.runners.python_function
from epsagon.common import EpsagonWarning
from epsagon.triggers.tencent_function import TencentFunctionTriggerFactory
from .. import constants


def _add_status_code(runner, return_value):
    """
    Tries to extract the status code from the return value and add it
    as a metadata field
    :param runner: Runner event to update
    :param return_value: The return value to extract from
    """
    if isinstance(return_value, collections.Mapping):
        status_code = return_value.get('statusCode')
        if status_code:
            runner.resource['metadata']['status_code'] = status_code


def tencent_function_wrapper(func):
    """Epsagon's Tencent SCF wrapper."""

    # avoid double instrumentation
    if getattr(func, '__instrumented__', False):
        return func

    @functools.wraps(func)
    def _tencent_function_wrapper(*args, **kwargs):
        """
        Generic SCF function wrapper
        """
        start_time = time.time()
        cold_start_duration = start_time - constants.COLD_START_TIME
        trace = epsagon.trace.trace_factory.get_or_create_trace()
        trace.prepare()

        try:
            event, context = args
        except ValueError:
            # This can happen when someone manually calls handler without
            # parameters / sends kwargs. In such case we ignore this trace.
            return func(*args, **kwargs)

        try:
            runner = epsagon.runners.tencent_function.TencentFunctionRunner(
                start_time,
                context
            )
            trace.set_runner(runner)
        # pylint: disable=W0703
        except Exception as exception:
            # Regress to python runner.
            warnings.warn(
                'SCF context is invalid, using simple python wrapper',
                EpsagonWarning
            )
            trace.add_exception(
                exception,
                traceback.format_exc()
            )
            return epsagon.wrappers.python_function.wrap_python_function(
                func,
                args,
                kwargs
            )

        if constants.COLD_START:
            runner.resource['metadata'][
                'tencent.scf.cold_start_duration'
            ] = cold_start_duration
        constants.COLD_START = False

        try:
            trace.add_event(
                TencentFunctionTriggerFactory.factory(
                    start_time,
                    event,
                    context,
                    runner
                )
            )
        # pylint: disable=W0703
        except Exception as exception:
            trace.add_exception(
                exception,
                traceback.format_exc(),
                additional_data={'event': event}
            )

        result = None
        try:
            result = func(*args, **kwargs)
            return result
        # pylint: disable=W0703
        except Exception as exception:
            runner.set_exception(
                exception,
                traceback.format_exc(),
                handled=False
            )
            raise
        finally:
            try:
                _add_status_code(runner, result)
                if not trace.metadata_only:
                    runner.resource['metadata']['tencent.scf.return_data'] = (
                        result
                    )
            # pylint: disable=W0703
            except Exception as exception:
                trace.add_exception(
                    exception,
                    traceback.format_exc(),
                )

            try:
                epsagon.trace.trace_factory.send_traces()
            # pylint: disable=W0703
            except Exception:
                epsagon.utils.print_debug('Failed to send SCF trace')

    _tencent_function_wrapper.__instrumented__ = True
    return _tencent_function_wrapper



"""
Triggers for Tencent Functions
"""

from __future__ import absolute_import
from uuid import uuid4
from epsagon.utils import add_data_if_needed
from ..event import BaseEvent


class BaseTencentFunctionTrigger(BaseEvent):
    """
    Represents base Tencent Function trigger
    """
    ORIGIN = 'trigger'


class JSONTrigger(BaseTencentFunctionTrigger):
    """
    Represents JSON trigger
    """
    RESOURCE_TYPE = 'json'

    def __init__(self, start_time, event, context, runner):
        """
        Initialize.
        :param start_time: event's start time (epoch)
        :param event: event dict from the entry point
        :param context: the context dict from the entry point
        :param runner: SCF runner event
        """

        super(JSONTrigger, self).__init__(start_time)

        self.event_id = 'trigger-{}'.format(str(uuid4()))

        self.resource['name'] = 'trigger-{}'.format(context['function_name'])
        self.resource['operation'] = self.RESOURCE_TYPE
        self.resource['metadata'] = {}

        add_data_if_needed(
            runner.resource['metadata'],
            'tencent.scf.trigger_data',
            event
        )


class TimerTrigger(BaseTencentFunctionTrigger):
    """
    Represents timer trigger
    """
    RESOURCE_TYPE = 'timer'

    # pylint: disable=W0613
    def __init__(self, start_time, event, context, runner):
        """
        Initialize.
        :param start_time: event's start time (epoch)
        :param event: event dict from the entry point
        :param context: the context dict from the entry point
        :param runner: SCF runner event
        """

        super(TimerTrigger, self).__init__(start_time)

        self.event_id = 'timer-{}'.format(str(uuid4()))
        self.resource['name'] = event['TriggerName']
        self.resource['operation'] = 'Timer'

        self.resource['metadata'] = {
            'tencent.timer.timestamp': event['Time'],
        }

        add_data_if_needed(
            self.resource['metadata'],
            'tencent.timer.message',
            event['Message']
        )


class COSTrigger(BaseTencentFunctionTrigger):
    """
    Represents Cloud Object Storage trigger
    """
    RESOURCE_TYPE = 'cos'

    # pylint: disable=W0613
    def __init__(self, start_time, event, context, runner):
        """
        Initialize.
        :param start_time: event's start time (epoch)
        :param event: event dict from the entry point
        :param context: the context dict from the entry point
        :param runner: SCF runner event
        """

        super(COSTrigger, self).__init__(start_time)
        record = event['Records'][0]
        self.event_id = record['cos']['cosObject']['meta']['x-cos-request-id']
        self.resource['name'] = '{}-{}'.format(
            record['cos']['cosBucket']['name'],
            record['cos']['cosBucket']['appid']
        )
        self.resource['operation'] = record['event']['eventName'].replace(
            'cos:',
            ''
        )

        self.resource['metadata'] = {
            'tencent.region': record['cos']['cosBucket']['cosRegion'],
            'tencent.app_id': record['cos']['cosBucket']['appid'],
            'tencent.cos.object_key': record['cos']['cosObject']['key'],
            'tencent.cos.object_size': record['cos']['cosObject']['size'],
            'tencent.cos.request_id': (
                record['cos']['cosObject']['meta']['x-cos-request-id']
            ),
        }


class CMQTrigger(BaseTencentFunctionTrigger):
    """
    Represents Cloud Message Queue trigger
    """
    RESOURCE_TYPE = 'cmq'

    # pylint: disable=W0613
    def __init__(self, start_time, event, context, runner):
        """
        Initialize.
        :param start_time: event's start time (epoch)
        :param event: event dict from the entry point
        :param context: the context dict from the entry point
        :param runner: SCF runner event
        """

        super(CMQTrigger, self).__init__(start_time)
        record = event['Records'][0]['CMQ']
        self.event_id = record['msgId']
        self.resource['name'] = record['topicName']
        self.resource['operation'] = 'consume'

        self.resource['metadata'] = {
            'tencent.cmq.message.id': record['msgId'],
            'tencent.cmq.message.tags': record['msgTag'],
            'tencent.cmq.request_id': record['requestId'],
            'tencent.cmq.subscription_name': record['subscriptionName'],
        }

        add_data_if_needed(
            self.resource['metadata'],
            'tencent.cmq.message.body',
            record['msgBody']
        )


class KafkaTrigger(BaseTencentFunctionTrigger):
    """
    Represents Kafka trigger
    """
    RESOURCE_TYPE = 'kafka'

    # pylint: disable=W0613
    def __init__(self, start_time, event, context, runner):
        """
        Initialize.
        :param start_time: event's start time (epoch)
        :param event: event dict from the entry point
        :param context: the context dict from the entry point
        :param runner: SCF runner event
        """

        super(KafkaTrigger, self).__init__(start_time)
        record = event['Records'][0]['Ckafka']
        self.event_id = record['msgKey']
        self.resource['name'] = record['topic']
        self.resource['operation'] = 'consume'

        self.resource['metadata'] = {
            'messaging.message.partition': record['partition'],
            'messaging.message.offset': record['offset'],
            'messaging.message.key': record['msgKey'],
        }

        add_data_if_needed(
            self.resource['metadata'],
            'messaging.message.body',
            record['msgBody']
        )


class APIGatewayTrigger(BaseTencentFunctionTrigger):
    """
    Represents API Gateway trigger
    """
    RESOURCE_TYPE = 'http'

    # pylint: disable=W0613
    def __init__(self, start_time, event, context, runner):
        """
        Initialize.
        :param start_time: event's start time (epoch)
        :param event: event dict from the entry point
        :param context: the context dict from the entry point
        :param runner: SCF runner event
        """

        super(APIGatewayTrigger, self).__init__(start_time)
        self.event_id = event['requestContext']['requestId']
        self.resource['name'] = event['headers']['Host']
        self.resource['operation'] = event['httpMethod']

        self.resource['metadata'] = {
            'http.route': event['requestContext']['path'],
            'http.request.path': event['path'],
            'tencent.api_gateway.request_id': (
                event['requestContext']['requestId']
            ),
            'tencent.api_gateway.stage': event['requestContext']['stage'],
        }

        add_data_if_needed(
            self.resource['metadata'],
            'http.request.headers',
            event['headers']
        )

        if event['body']:
            add_data_if_needed(
                self.resource['metadata'],
                'http.request.body',
                event['body']
            )

        if event['pathParameters']:
            add_data_if_needed(
                self.resource['metadata'],
                'http.request.path_params',
                event['pathParameters']
            )

        if event['queryString']:
            add_data_if_needed(
                self.resource['metadata'],
                'http.request.query',
                event['queryString']
            )


class TencentFunctionTriggerFactory(object):
    """
    Represents a Tencent Function Trigger Factory.
    """

    @staticmethod
    def factory(start_time, event, context, runner):
        """
        Creates trigger event object.
        :param start_time: event's start time (epoch)
        :param event: event dict from the entry point
        :param context: the context dict from the entry point
        :param runner: SCF runner event
        :return: Event
        """
        trigger_service = JSONTrigger

        if event.get('Type', None) == 'Timer':
            trigger_service = TimerTrigger
        elif 'httpMethod' in event:
            trigger_service = APIGatewayTrigger
        elif 'Records' in event:
            if 'cos' in event['Records'][0]:
                trigger_service = COSTrigger
            elif 'CMQ' in event['Records'][0]:
                trigger_service = CMQTrigger
            elif 'Ckafka' in event['Records'][0]:
                trigger_service = KafkaTrigger

        return trigger_service(
            start_time,
            event,
            context,
            runner
        )



"""
Celery events module
"""

from __future__ import absolute_import
import time
import functools
import traceback
from uuid import uuid4
from importlib import import_module
from ..trace import trace_factory
from ..event import BaseEvent
from ..utils import add_data_if_needed
from ..runners.celery import CeleryRunner

# A map of all active events and pending runners. The key is the `{sender}-{id}`
ACTIVE_EVENTS = {}


class CeleryEvent(BaseEvent):
    """
    Represents Celery event.
    """

    ORIGIN = 'celery'
    RESOURCE_TYPE = 'celery'
    OPERATION = 'publish'
    DRIVER_MAPPING = {
        'amqp': 'rabbitmq',
        'redis': 'redis',
    }

    def __init__(self, *args, **kwargs):  # pylint: disable=unused-argument
        """
        Initialize.
        """
        super(CeleryEvent, self).__init__(time.time())

        self.event_id = str(uuid4())
        self.resource['name'] = kwargs.get('sender', '')
        self.resource['operation'] = self.OPERATION

        body = kwargs.get('body', [None])[0]
        if body:
            # Body comes in tuple which is not serializable
            # so we change it to list
            body = list(body)

        app_conn = import_module('celery').current_app.connection()
        headers = kwargs.get('headers', {})

        self.resource['metadata'] = {
            'origin': headers.get('origin', ''),
            'retries': headers.get('retries', ''),
            'id': headers.get('id', ''),
            'routing_key': kwargs.get('routing_key', ''),
            'hostname': app_conn.hostname,
            'virtual_host': app_conn.virtual_host,
            'driver': app_conn.transport.driver_type,
        }

        # Check if this is a known driver to update the resource details
        driver_map = self.DRIVER_MAPPING.get(app_conn.transport.driver_type)
        if driver_map:
            self.resource['name'] = app_conn.hostname
            self.resource['type'] = '{}_{}'.format(
                self.RESOURCE_TYPE,
                driver_map
            )

        add_data_if_needed(
            self.resource['metadata'],
            'body',
            body
        )


def get_event_key(*args, **kwargs):  # pylint: disable=unused-argument
    """
    Returns the event key to get from ACTIVE_EVENTS.
    :return: `sender-id` string or None if input does not match
    """
    if 'task_id' in kwargs:
        # Comes from pre and post run, and failure signals
        event_id = kwargs.get('task_id', '')
        sender = kwargs.get('sender').name if kwargs.get('sender') else ''
    elif 'request' in kwargs:
        # Comes from retry signals
        event_id = kwargs.get('request', {}).get('id', '')
        sender = kwargs.get('sender').name if kwargs.get('sender') else ''
    else:
        # Comes from before and after publish signals
        event_id = kwargs.get('headers', {}).get('id', '')
        sender = kwargs.get('sender', '')

    if event_id == '' or sender == '':
        return None

    return '{}-{}'.format(sender, event_id)


def signal_wrapper(func):
    """
    Signal fail-safe wrapper
    :param func: original signal
    :return: fail-safe signal function
    """
    @functools.wraps(func)
    def _signal_wrapper(*args, **kwargs):
        """
        Wraps original signal wrapper with try
        """
        try:
            return func(*args, **kwargs)
        except Exception as exception:  # pylint: disable=broad-except
            event = ACTIVE_EVENTS.get(get_event_key(*args, **kwargs))
            if event:
                trace_factory.add_exception(
                    exception,
                    traceback.format_exc()
                )

    return _signal_wrapper


# -----------------------
#  Celery publish events
# -----------------------

@signal_wrapper
def wrap_before_publish(*args, **kwargs):
    """
    Wraps before publish signal (task.delay or task.apply_async)
    """
    event_key = get_event_key(*args, **kwargs)
    if event_key:
        ACTIVE_EVENTS[event_key] = CeleryEvent(*args, **kwargs)


@signal_wrapper
def wrap_after_publish(*args, **kwargs):
    """
    Wraps after publish signal (task.delay or task.apply_async)
    """
    event_key = get_event_key(*args, **kwargs)
    event = ACTIVE_EVENTS.pop(event_key, None)
    if event:
        event.terminate()
        trace_factory.add_event(event)


# ----------------------
#  Celery runner events
# ----------------------

@signal_wrapper
def wrap_prerun(*args, **kwargs):
    """
    Wraps pre-run signal of worker
    """
    event_key = get_event_key(*args, **kwargs)
    if event_key:
        trace_factory.get_or_create_trace(unique_id=event_key)
        runner = CeleryRunner(*args, **kwargs)
        ACTIVE_EVENTS[event_key] = runner
        trace_factory.set_runner(runner)


@signal_wrapper
def wrap_postrun(*args, **kwargs):
    """
    Wraps post-run signal of worker
    """
    event_key = get_event_key(*args, **kwargs)
    event = ACTIVE_EVENTS.pop(event_key, None)
    if event:
        event.terminate()
        trace_factory.send_traces()


@signal_wrapper
def wrap_retry(*args, **kwargs):
    """
    Wraps retry signal of worker
    """
    event = ACTIVE_EVENTS.get(get_event_key(*args, **kwargs))
    if event:
        event.set_retry(kwargs.get('request', {}).get('retries', 0))


@signal_wrapper
def wrap_failure(*args, **kwargs):
    """
    Wraps failure signal of worker
    """
    event = ACTIVE_EVENTS.get(get_event_key(*args, **kwargs))
    if event:
        event.set_exception(
            kwargs.get('exception', Exception),
            traceback.format_exc()
        )



"""
Runner for Tencent SCF.
"""

from __future__ import absolute_import
from ..event import BaseEvent
from .. import constants
from ..common import ErrorCode


class TencentFunctionRunner(BaseEvent):
    """
    Represents Tencent SCF event runner.
    """
    ORIGIN = 'runner'
    RESOURCE_TYPE = 'tencent_function'
    OPERATION = 'invoke'

    def __init__(self, start_time, context):
        """
        Initialize.
        :param start_time: event's start time (epoch)
        :param context: SCF's context (passed from entry point)
        """
        super(TencentFunctionRunner, self).__init__(start_time)

        # Creating a unique ID for local runs.
        self.event_id = context['request_id']
        self.resource['name'] = context['function_name']
        self.resource['operation'] = self.OPERATION

        self.resource['metadata'].update({
            'tencent.scf.version': context['function_version'],
            'tencent.scf.memory': context['memory_limit_in_mb'],
            'tencent.scf.cold_start': constants.COLD_START,
            'tencent.namespace': context['namespace'],
            'tencent.uin': context['tencentcloud_uin'],
            'tencent.app_id': context['tencentcloud_appid'],
            'tencent.region': context['tencentcloud_region'],
        })

    def set_timeout(self):
        """
        Sets timeout error code.
        :return: None
        """
        # Don't override exceptions
        if self.error_code != ErrorCode.EXCEPTION:
            self.error_code = ErrorCode.TIMEOUT



"""
redis events module.
"""

from __future__ import absolute_import
from uuid import uuid4
import traceback

from ..event import BaseEvent
from ..trace import trace_factory
from ..utils import add_data_if_needed


MAX_VALUE_SIZE = 1024
MAX_CMD_PIPELINE = 10


def _parse_redis_cmd(cmd_args):
    """
    Parse redis cmd to operation, arguments
    :param cmd_args: command arguments
    :return: operation, key
    """
    key = ''
    operation = str(cmd_args[0])[:MAX_VALUE_SIZE]
    if len(cmd_args) > 1:
        key = str(cmd_args[1])[:MAX_VALUE_SIZE]
    return operation, key


def _parse_redis_connection(connection):
    """
    Parse redis connection to host, db
    :param connection: redis connection
    :return: host, db
    """
    connection_kwargs = connection.connection_pool.connection_kwargs
    return (
        connection_kwargs.get('host', 'local'),
        connection_kwargs.get('port'),
        connection_kwargs.get('db', '0')
    )


class BaseRedisEvent(BaseEvent):
    """
    Represents base redis event.
    """
    ORIGIN = 'redis'
    RESOURCE_TYPE = 'redis'

    # pylint: disable=W0613
    def __init__(self, wrapped, instance, args, kwargs, start_time, response,
                 exception):
        """
        Initialize the redis event
        :param wrapped: wrapt's wrapped
        :param instance: wrapt's instance
        :param args: wrapt's args
        :param kwargs: wrapt's kwargs
        :param start_time: Start timestamp (epoch)
        :param response: response data
        :param exception: Exception (if happened)
        """

        super(BaseRedisEvent, self).__init__(start_time)

        self.event_id = 'redis-{}'.format(str(uuid4()))

        host, port, db = _parse_redis_connection(instance)

        self.resource['name'] = host
        self.resource['metadata'] = {
            'Redis Host': host,
            'Redis Port': port,
            'Redis DB Index': db
        }

        if exception is not None:
            self.set_exception(exception, traceback.format_exc())


class RedisSingleExecutionEvent(BaseRedisEvent):
    """
    Represents single execution redis event.
    """

    ORIGIN = 'redis'
    RESOURCE_TYPE = 'redis'

    # pylint: disable=W0613
    def __init__(self, wrapped, instance, args, kwargs, start_time, response,
                 exception):
        """
        Initialize.
        :param wrapped: wrapt's wrapped
        :param instance: wrapt's instance
        :param args: wrapt's args
        :param kwargs: wrapt's kwargs
        :param start_time: Start timestamp (epoch)
        :param response: response data
        :param exception: Exception (if happened)
        """

        super(RedisSingleExecutionEvent, self).__init__(
            wrapped, instance, args, kwargs, start_time, response, exception
        )

        self.event_id = 'redis-{}'.format(str(uuid4()))

        operation, key = _parse_redis_cmd(args)
        self.resource['operation'] = operation
        self.resource['metadata']['Redis Key'] = key

        if response:
            add_data_if_needed(
                self.resource['metadata'],
                'redis.response',
                response
            )


class RedisMultiExecutionEvent(BaseRedisEvent):
    """
    Represents base redis event.
    """

    ORIGIN = 'redis'
    RESOURCE_TYPE = 'redis'

    # pylint: disable=W0613
    def __init__(self, wrapped, instance, args, kwargs, start_time, response,
                 exception, command_stack):
        """
        Initialize.
        :param wrapped: wrapt's wrapped
        :param instance: wrapt's instance
        :param args: wrapt's args
        :param kwargs: wrapt's kwargs
        :param start_time: Start timestamp (epoch)
        :param response: response data
        :param exception: Exception (if happened)
        """

        super(RedisMultiExecutionEvent, self).__init__(
            wrapped, instance, args, kwargs, start_time, response, exception
        )

        self.resource['operation'] = 'Pipeline'

        operations = [
            '{} {}'.format(op, key) for op, key in [
                _parse_redis_cmd(cmd_args)
                for cmd_args, _ in command_stack
            ]
        ]
        self.resource['metadata']['Stack Count'] = len(operations)
        self.resource['metadata']['Actions'] = operations[:MAX_CMD_PIPELINE]


class RedisMultiEventFactory(object):
    """
    Factory class, generates Redis multi-execution event.
    """
    LAST_STACK = []

    @staticmethod
    def create_event(wrapped, instance, args, kwargs, start_time, response,
                     exception):
        """
        Create a Redis event.
        :param wrapped:
        :param instance:
        :param args:
        :param kwargs:
        :param start_time:
        :param response:
        :param exception:
        :return:
        """
        event = RedisMultiExecutionEvent(
            wrapped,
            instance,
            args,
            kwargs,
            start_time,
            response,
            exception,
            RedisMultiEventFactory.LAST_STACK
        )
        trace_factory.add_event(event)


class RedisSingleEventFactory(object):
    """
    Factory class, generates Redis event.
    """
    @staticmethod
    def create_event(wrapped, instance, args, kwargs, start_time, response,
                     exception):
        """
        Create a Redis event.
        :param wrapped:
        :param instance:
        :param args:
        :param kwargs:
        :param start_time:
        :param response:
        :param exception:
        :return:
        """
        event = RedisSingleExecutionEvent(
            wrapped,
            instance,
            args,
            kwargs,
            start_time,
            response,
            exception
        )
        trace_factory.add_event(event)



"""trace transport layers"""

import os
import base64
import logging
import json
import urllib3
from epsagon.constants import SEND_TIMEOUT
from epsagon.trace_encoder import TraceEncoder


def to_json(obj):
    return json.dumps(obj, cls=TraceEncoder, ensure_ascii=True)


class NoneTransport(object):
    @classmethod
    def send(cls, _):
        logging.error('trace sent using NoneTransport, configure a transport')


class LogTransport(object):
    """ send traces by logging them """

    @staticmethod
    def send(trace):
        trace_json = to_json(trace.to_dict())
        trace_message = base64.b64encode(
            trace_json.encode('utf-8')
        ).decode('utf-8')

        # pylint: disable=superfluous-parens
        print('EPSAGON_TRACE: {}'.format(trace_message))


class HTTPTransport(object):
    """ send traces using http request """

    def __init__(self, dest, token):
        self.dest = dest
        self.token = token
        self.timeout = SEND_TIMEOUT
        self.session = urllib3.PoolManager(
            cert_reqs='CERT_REQUIRED',
            ca_certs=os.path.join(os.path.dirname(__file__), 'cacert.pem'),
            headers={
                'Authorization': 'Bearer {}'.format(self.token),
                'Content-Type': 'application/json'
            },
            # max size of reusable connections
            maxsize=5
        )

    def send(self, trace):
        self.session.request(
            'POST',
            self.dest,
            body=to_json(trace.to_dict()),
            timeout=self.timeout
        )



"""
Common tests helpers
"""
import mock
import requests
from threading import Thread


def get_tracer_patch_kwargs():
    return {
        'metadata_only': False,
        'disable_timeout_send': False,
        'prepare': mock.MagicMock(),
        'send_traces': mock.MagicMock(),
        'events': [],
        'add_event': mock.MagicMock(),
        'add_exception': mock.MagicMock(),
        'set_runner': mock.MagicMock()
    }


def _send_get_request(target_url, results):
    """
    Sends a get requests to a given target URL string
    :return: the given target url
    """
    requests.get(target_url)
    results.append(target_url)


def multiple_threads_handler(threads_count=3):
    """
    Invokes `threads_count` new threads, each performs an HTTP get request.
    Waits for all threads and validates a result has been returned from each
    thread.
    """
    threads = []
    results = []
    for i in range(threads_count):
        thread = Thread(target = _send_get_request, args = ("http://google.com", results))
        thread.start()
        threads.append(thread)
    for thread in threads:
        thread.join()
    assert len(results) == threads_count



"""
Azure sdk patcher module.
"""

from __future__ import absolute_import
import wrapt
from epsagon.modules.general_wrapper import wrapper
from ..events.azure import AzureEventFactory


def _wrapper(wrapped, instance, args, kwargs):
    """
    General wrapper for Azure sdk instrumentation.
    :param wrapped: wrapt's wrapped
    :param instance: wrapt's instance
    :param args: wrapt's args
    :param kwargs: wrapt's kwargs
    :return: None
    """
    return wrapper(AzureEventFactory, wrapped, instance, args, kwargs)


def patch():
    """
    Patch module.
    :return: None
    """
    wrapt.wrap_function_wrapper(
        'azure.cosmos.container',
        'ContainerProxy.delete_item',
        _wrapper
    )
    wrapt.wrap_function_wrapper(
        'azure.cosmos.container',
        'ContainerProxy.upsert_item',
        _wrapper
    )
    wrapt.wrap_function_wrapper(
        'azure.cosmos.container',
        'ContainerProxy.query_items',
        _wrapper
    )



"""
Wrapper for AWS Lambda in Chalice environment.
"""

from __future__ import absolute_import
from .aws_lambda import lambda_wrapper


class ChaliceWrapper(object):
    """
    Class handles wrapping Chalice app.
    In call we expect an invocation to come, and in getattr we allow `app.attr`
    calls.
    """
    def __init__(self, app):
        self._app = app

    def __getattr__(self, item):
        return getattr(self._app, item)

    def __setattr__(self, name, value):
        if name == '__class__' and value.__name__ == 'LocalChalice':
            # In local runs, the class is being changed to `LocalChalice`,
            # So we do that to preserve the same behaviour
            value.__getattr__ = self.__getattr__
        super(ChaliceWrapper, self).__setattr__(name, value)

    def __call__(self, *args, **kwargs):
        return lambda_wrapper(self._app)(*args, **kwargs)


def chalice_wrapper(app):
    """
    Chalice wrapper
    :param app: Chalice app
    :return: ChaliceWrapper
    """
    return ChaliceWrapper(app)



"""
pymongo events module.
"""

from __future__ import absolute_import
from uuid import uuid4
import traceback
from past.builtins import range

from epsagon.utils import add_data_if_needed
from ..event import BaseEvent
from ..trace import trace_factory


class PyMongoEvent(BaseEvent):
    """
    Represents base pymongo event.
    """

    ORIGIN = 'pymongo'
    RESOURCE_TYPE = 'pymongo'

    #pylint: disable=W0613
    def __init__(self, wrapped, instance, args, kwargs, start_time, response,
                 exception):
        """
        Initialize.
        :param wrapped: wrapt's wrapped
        :param instance: wrapt's instance
        :param args: wrapt's args
        :param kwargs: wrapt's kwargs
        :param start_time: Start timestamp (epoch)
        :param response: response data
        :param exception: Exception (if happened)
        """

        super(PyMongoEvent, self).__init__(start_time)

        documents = args[0]

        self.event_id = 'mongo-{}'.format(str(uuid4()))
        self.resource['name'] = instance.full_name

        self.resource['operation'] = \
            'insert_many' if isinstance(documents, list) else 'insert_one'
        address = list(getattr(
            instance.database.client,
            '_topology_settings'
        ).seeds)[0]

        if self.resource['operation'] == 'insert_one':
            documents = [documents]

        self.resource['metadata'] = {
            'DB URL': ':'.join([str(x) for x in address]),
            'DB Name': str(instance.database.name),
            'Collection Name': str(instance.name),
        }
        add_data_if_needed(self.resource['metadata'], 'Items', documents)

        if response is not None:
            self.update_response(response)

        if exception is not None:
            self.set_exception(exception, traceback.format_exc())

    def update_response(self, response):
        """
        Adds response data to event.
        :param response: Response from botocore
        :return: None
        """
        for i in range(len(self.resource['metadata'].get('items', []))):
            self.resource['metadata']['items'][i]['_id'] = str(
                self.resource['metadata']['items'][i]['_id']
            )

        if self.resource['operation'] == 'insert_many':
            self.resource['metadata']['inserted_ids'] = \
                [str(x) for x in response.inserted_ids]
        elif self.resource['operation'] == 'insert_one':
            self.resource['metadata']['inserted_ids'] = \
                [str(response.inserted_id)]


class PyMongoEventFactory(object):
    """
    Factory class, generates MongoDB event.
    """

    @staticmethod
    def create_event(wrapped, instance, args, kwargs, start_time, response,
                     exception):
        """
        Create a PyMongo event.
        :param wrapped:
        :param instance:
        :param args:
        :param kwargs:
        :param start_time:
        :param response:
        :param exception:
        :return:
        """
        event = PyMongoEvent(
            wrapped,
            instance,
            args,
            kwargs,
            start_time,
            response,
            exception
        )
        trace_factory.add_event(event)



"""
Wrapper for Azure Function.
"""

from __future__ import absolute_import
import time
import traceback
import warnings
import functools
from ..trace import trace_factory
from ..runners.azure_function import AzureFunctionRunner
from ..triggers.azure_function import AzureTriggerFactory
from ..common import EpsagonWarning


def azure_wrapper(func):
    """Epsagon's Azure Function wrapper."""

    @functools.wraps(func)
    def _azure_wrapper(*args, **kwargs):
        """
        general Azure function wrapper
        """
        trace = trace_factory.get_or_create_trace()
        trace.prepare()

        context = kwargs.get('context')
        if not context:
            return func(*args, **kwargs)

        # Create Runner
        try:
            runner = AzureFunctionRunner(time.time(), context)
            trace.set_runner(runner)
        except Exception as exception:  # pylint: disable=broad-except
            warnings.warn(
                'Could not create Azure Function runner: {}'.format(exception),
                EpsagonWarning
            )
            return func(*args, **kwargs)

        result = None
        try:
            result = func(*args, **kwargs)
        except Exception as exception:
            runner.set_exception(exception, traceback.format_exc())
            raise
        finally:
            runner.terminate()

        # Create Trigger
        try:
            azure_trigger = AzureTriggerFactory.factory(
                time.time(),
                kwargs,
                result
            )
            if azure_trigger:
                trace.add_event(azure_trigger)
        except Exception as exception:  # pylint: disable=broad-except
            warnings.warn(
                'Could not create Azure Function trigger: {}'.format(exception),
                EpsagonWarning
            )
            return func(*args, **kwargs)

        trace_factory.send_traces()
        return result

    return _azure_wrapper



"""
Wrapper for Python Flask.
"""

from __future__ import absolute_import
import traceback
import time
import warnings

from flask import request
import epsagon.trace
import epsagon.triggers.http
import epsagon.runners.flask
from epsagon.common import EpsagonWarning
from epsagon.utils import collect_container_metadata,\
    get_traceback_data_from_exception
from ..http_filters import ignore_request, \
    is_ignored_endpoint, add_ignored_endpoints


class FlaskWrapper(object):
    """
    Wraps Flask wsgi application.
    """
    EPSAGON_MARKER = '_epsagon_wrapper'

    def __init__(self, app, ignored_endpoints=None):
        """
        WSGI app wrapper for flask application.
        :param app: the :class:`flask.Flask` application object.
        :param ignored_endpoints: endpoint paths to ignore.
        """
        # Wrapping app only once
        if getattr(app, self.EPSAGON_MARKER, False):
            return
        setattr(app, self.EPSAGON_MARKER, True)
        self.app = app
        if ignored_endpoints:
            if isinstance(ignored_endpoints, list):
                add_ignored_endpoints(ignored_endpoints)
            elif isinstance(ignored_endpoints, str):
                add_ignored_endpoints([ignored_endpoints])

        # Override request handling.
        self.app.before_request(self._before_request)
        self.app.after_request(self._after_request)
        self.app.teardown_request(self._teardown_request)

        # Whether we ignore this request or not.
        self.ignored_request = False
        epsagon.trace.trace_factory.switch_to_multiple_traces()

    def _before_request(self):
        """
        Runs when new request comes in.
        :return: None.
        """
        # Ignoring non relevant content types.
        self.ignored_request = ignore_request('', request.path.lower())

        if self.ignored_request:
            return

        trace = epsagon.trace.trace_factory.get_or_create_trace()
        trace.prepare()

        # Create flask runner with current request.
        try:
            runner = epsagon.runners.flask.FlaskRunner(
                time.time(),
                self.app,
                request
            )
            trace.set_runner(runner)

            # Collect metadata in case this is a container.
            collect_container_metadata(runner.resource['metadata'])

        # pylint: disable=W0703
        except Exception as exception:
            # Regress to python runner.
            warnings.warn('Could not extract request', EpsagonWarning)
            trace.add_exception(
                exception,
                traceback.format_exc()
            )

        # Extract HTTP trigger data.
        try:
            trigger = epsagon.triggers.http.HTTPTriggerFactory.factory(
                    time.time(),
                    request
                )
            if trigger:
                trace.add_event(trigger)
        # pylint: disable=W0703
        except Exception as exception:
            trace.add_exception(
                exception,
                traceback.format_exc(),
            )

    def _after_request(self, response):
        """
        Runs after first process of response.
        :param response: The current Response object.
        :return: Response.
        """
        if self.ignored_request:
            return response

        # Ignoring non relevant content types.
        if ignore_request(response.content_type.lower(), ''):
            self.ignored_request = True
            return response

        trace = epsagon.trace.trace_factory.get_or_create_trace()

        if trace.runner:
            trace.runner.update_response(response)
        return response

    def _teardown_request(self, exception):
        """
        Runs at the end of the request. Exception will be passed if happens.
        If no flask url rule exists for a request, then the request trace
        will be passed.
        :param exception: Exception (or None).
        :return: None.
        """
        if self.ignored_request:
            return
        trace = epsagon.trace.trace_factory.get_or_create_trace()
        if exception and trace.runner:
            traceback_data = get_traceback_data_from_exception(exception)
            trace.runner.set_exception(exception, traceback_data)
        # Ignoring endpoint, only if no error happened.
        if (not exception and
            request.url_rule and
            is_ignored_endpoint(request.url_rule.rule)
        ):
            return

        epsagon.trace.trace_factory.send_traces()



"""
botocore events module.
"""

# pylint: disable=C0302
from __future__ import absolute_import

import hashlib
import traceback
from importlib import import_module
import json
from epsagon.constants import STEP_DICT_NAME
from ..trace import trace_factory
from ..event import BaseEvent
from ..utils import add_data_if_needed, add_metadata_from_dict

# Conditionally importing boto3
ClientError = Exception  # pylint: disable=invalid-name
TypeDeserializer = None  # pylint: disable=invalid-name
ConditionExpressionBuilder = None  # pylint: disable=invalid-name
try:
    ClientError = (  # pylint: disable=invalid-name
        import_module('botocore.exceptions').ClientError
    )
    TypeDeserializer = (  # pylint: disable=invalid-name
        import_module('boto3.dynamodb.types').TypeDeserializer
    )
    ConditionExpressionBuilder = (  # pylint: disable=invalid-name
        import_module('boto3.dynamodb.conditions').ConditionExpressionBuilder
    )
except ImportError:
    pass


# pylint: disable=W0613
def empty_func(*args):
    """
    A dummy function.
    :return: None
    """


class BotocoreEvent(BaseEvent):
    """
    Represents base botocore event.
    """

    ORIGIN = 'botocore'
    RESOURCE_TYPE = 'botocore'
    RESPONSE_TO_FUNC = {}
    OPERATION_TO_FUNC = {}

    # pylint: disable=W0613
    def __init__(self, wrapped, instance, args, kwargs, start_time, response,
                 exception):
        """
        Initialize.
        :param wrapped: wrapt's wrapped
        :param instance: wrapt's instance
        :param args: wrapt's args
        :param kwargs: wrapt's kwargs
        :param start_time: Start timestamp (epoch)
        :param response: response data
        :param exception: Exception (if happened)
        """
        super(BotocoreEvent, self).__init__(start_time)

        event_operation, _ = args
        self.resource['operation'] = str(event_operation)
        self.resource['name'] = ''

        self.resource['metadata'] = {
            'region': instance.meta.region_name,
        }

        if response is not None:
            self.update_response(response)

        if exception is not None:
            self.set_exception(exception, traceback.format_exc())

    def set_exception(
            self,
            exception,
            traceback_data,
            handled=True,
            from_logs=False
    ):
        """
        see {Event.set_exception}
        """
        super(BotocoreEvent, self).set_exception(
            exception,
            traceback_data,
            handled=handled,
            from_logs=from_logs
        )

        # Specific handling for botocore errors
        if (
            isinstance(exception, ClientError) and
            'ResponseMetadata' in exception.response
        ):
            self.event_id = exception.response['ResponseMetadata']['RequestId']
            botocore_error = exception.response['Error']
            self.resource['metadata']['botocore_error'] = True
            self.resource['metadata']['error_code'] = \
                str(botocore_error.get('Code', ''))
            self.resource['metadata']['error_message'] = \
                str(botocore_error.get('Message', ''))
            self.resource['metadata']['error_type'] = \
                str(botocore_error.get('Type', ''))

    def update_response(self, response):
        """
        Adds response data to event.
        :param response: Response from botocore
        :return: None
        """
        event_id = response['ResponseMetadata'].get('RequestId')
        if event_id:
            self.event_id = event_id
        self.resource['metadata']['Retry Attempts'] = \
            response['ResponseMetadata']['RetryAttempts']
        self.resource['metadata']['Status Code'] = \
            response['ResponseMetadata']['HTTPStatusCode']


class BotocoreCloudWatchEvent(BotocoreEvent):
    """
    Represents cloudwatch events (eventbridge) botocore event.
    """

    RESOURCE_TYPE = 'eventbridge'
    RESOURCE_TYPE_UPDATE = 'events'

    def __init__(
        self, wrapped, instance, args, kwargs, start_time, response, exception
    ):
        """
        Initialize.
        :param wrapped: wrapt's wrapped
        :param instance: wrapt's instance
        :param args: wrapt's args
        :param kwargs: wrapt's kwargs
        :param start_time: Start timestamp (epoch)
        :param response: response data
        :param exception: Exception (if happened)
        """

        super(BotocoreCloudWatchEvent, self).__init__(
            wrapped, instance, args, kwargs, start_time, response, exception
        )
        _, request_data = args
        entries = request_data.get('Entries')[0] if request_data.get('Entries')\
            else {}
        self.resource['name'] = entries.get('EventBusName', 'CloudWatch Events')
        if self.resource['operation'] == 'PutEvents':
            if 'DetailType' in entries:
                self.resource['metadata']['aws.cloudwatch.detail_type'] = \
                    entries[
                        'DetailType'
                    ]
            if 'Resources' in entries:
                self.resource['metadata']['aws.cloudwatch.resources'] = \
                    entries[
                        'Resources'
                    ]
            if 'Source' in entries:
                self.resource['metadata']['aws.cloudwatch.source'] = entries[
                    'Source'
                ]
            if 'Detail' in entries:
                add_data_if_needed(
                    self.resource['metadata'],
                    'aws.cloudwatch.detail',
                    entries['Detail']
                )

        self.resource['type'] = self.RESOURCE_TYPE_UPDATE

    def update_response(self, response):
        """
        Adds response data to event.
        :param response: Response from botocore
        :return: None
        """
        super(BotocoreCloudWatchEvent, self).update_response(response)
        if self.resource['operation'] == 'PutEvents' and 'Entries' in response \
                and response['Entries']:
            self.resource['metadata']['aws.cloudwatch.event_id'] = \
                response['Entries'][0]['EventId']

class BotocoreS3Event(BotocoreEvent):
    """
    Represents s3 botocore event.
    """

    RESOURCE_TYPE = 's3'

    def __init__(self, wrapped, instance, args, kwargs, start_time, response,
                 exception):
        """
        Initialize.
        :param wrapped: wrapt's wrapped
        :param instance: wrapt's instance
        :param args: wrapt's args
        :param kwargs: wrapt's kwargs
        :param start_time: Start timestamp (epoch)
        :param response: response data
        :param exception: Exception (if happened)
        """
        super(BotocoreS3Event, self).__init__(
            wrapped,
            instance,
            args,
            kwargs,
            start_time,
            response,
            exception
        )

        _, request_data = args
        self.resource['name'] = request_data['Bucket']

        if self.resource['operation'] in \
                ['HeadObject', 'GetObject', 'PutObject']:
            self.resource['metadata']['key'] = request_data['Key']
        elif self.resource['operation'] == 'CopyObject':
            self.resource['metadata']['source'] = request_data['CopySource']
            self.resource['metadata']['destination'] = request_data['Key']

    def update_response(self, response):
        """
        Adds response data to event.
        :param response: Response from botocore
        :return: None
        """
        super(BotocoreS3Event, self).update_response(response)

        if self.resource['operation'] == 'ListObjects':
            files = [
                [str(x['Key']).strip('"'), x['Size'], x['ETag']]
                for x in response.get('Contents', [])
            ]
            add_data_if_needed(self.resource['metadata'], 'files', files)

        elif self.resource['operation'] == 'PutObject':
            self.resource['metadata']['etag'] = response['ETag'].strip('"')
        elif self.resource['operation'] == 'HeadObject':
            self.resource['metadata']['etag'] = response['ETag'].strip('"')
            self.resource['metadata']['file_size'] = response['ContentLength']
            self.resource['metadata']['last_modified'] = \
                response['LastModified'].strftime('%s')
        elif self.resource['operation'] == 'GetObject':
            self.resource['metadata']['etag'] = response['ETag'].strip('"')
            self.resource['metadata']['file_size'] = response['ContentLength']
            self.resource['metadata']['last_modified'] = \
                response['LastModified'].strftime('%s')


class BotocoreKinesisEvent(BotocoreEvent):
    """
    Represents kinesis botocore event.
    """
    RESOURCE_TYPE = 'kinesis'

    def __init__(self, wrapped, instance, args, kwargs, start_time, response,
                 exception):
        """
        Initialize.
        :param wrapped: wrapt's wrapped
        :param instance: wrapt's instance
        :param args: wrapt's args
        :param kwargs: wrapt's kwargs
        :param start_time: Start timestamp (epoch)
        :param response: response data
        :param exception: Exception (if happened)
        """

        super(BotocoreKinesisEvent, self).__init__(
            wrapped,
            instance,
            args,
            kwargs,
            start_time,
            response,
            exception
        )

        _, request_data = args
        self.resource['name'] = request_data['StreamName']
        if 'Data' in request_data:
            add_data_if_needed(
                self.resource['metadata'],
                'data',
                request_data['Data']
            )
        if 'PartitionKey' in request_data:
            self.resource['metadata']['partition_key'] = \
                request_data['PartitionKey']

    def update_response(self, response):
        """
        Adds response data to event.
        :param response: Response from botocore
        :return: None
        """
        super(BotocoreKinesisEvent, self).update_response(response)

        if self.resource['operation'] == 'PutRecord':
            self.resource['metadata']['shard_id'] = response['ShardId']
            self.resource['metadata']['sequence_number'] = \
                response['SequenceNumber']
        if self.resource['operation'] == 'PutRecords':
            self.resource['metadata']['failed_record_count'] = \
                response['FailedRecordCount']
            self.resource['metadata']['shard_id'] = \
                response['Records'][0]['ShardId']
            self.resource['metadata']['sequence_number'] = \
                response['Records'][0]['SequenceNumber']
            self.resource['metadata']['record_count'] = \
                len(response['Records'])


class BotocoreSNSEvent(BotocoreEvent):
    """
    Represents SNS botocore event.
    """
    RESOURCE_TYPE = 'sns'

    def __init__(self, wrapped, instance, args, kwargs, start_time, response,
                 exception):
        """
        Initialize.
        :param wrapped: wrapt's wrapped
        :param instance: wrapt's instance
        :param args: wrapt's args
        :param kwargs: wrapt's kwargs
        :param start_time: Start timestamp (epoch)
        :param response: response data
        :param exception: Exception (if happened)
        """
        super(BotocoreSNSEvent, self).__init__(
            wrapped,
            instance,
            args,
            kwargs,
            start_time,
            response,
            exception
        )
        _, request_data = args

        if self.resource['operation'] == 'CreateTopic':
            self.resource['name'] = request_data.get('Name', 'N/A')
        else:
            arn = request_data.get(
                'TopicArn',
                request_data.get('TargetArn', 'N/A')
            )
            self.resource['name'] = arn.split(':')[-1]

        self._process_request_data(request_data)

    def _process_request_data(self, request_data):
        """
        Process the SNS message request data - adding relevant message data
        :param request_data: the request_data to process
        """
        message_fields_description = {
            'Message': 'Notification Message',
            'MessageAttributes': 'Notification Message Attributes',
        }
        for field, description in message_fields_description.items():
            if field in request_data:
                add_data_if_needed(
                    self.resource['metadata'],
                    description,
                    request_data[field]
                )

        header_data = {
            key: value for key, value in request_data.items()
            if key not in message_fields_description
        }
        add_data_if_needed(
            self.resource['metadata'],
            'Notification Message Headers',
            header_data
        )

    def update_response(self, response):
        """
        Adds response data to event.
        :param response: Response from botocore
        :return: None
        """

        super(BotocoreSNSEvent, self).update_response(response)

        if self.resource['operation'] == 'Publish':
            self.resource['metadata']['message_id'] = response['MessageId']


class BotocoreSQSEvent(BotocoreEvent):
    """
    Represents SQS botocore event.
    """
    RESOURCE_TYPE = 'sqs'

    def __init__(self, wrapped, instance, args, kwargs, start_time, response,
                 exception):
        """
        Initialize.
        :param wrapped: wrapt's wrapped
        :param instance: wrapt's instance
        :param args: wrapt's args
        :param kwargs: wrapt's kwargs
        :param start_time: Start timestamp (epoch)
        :param response: response data
        :param exception: Exception (if happened)
        """
        self.RESPONSE_TO_FUNC.update({
            'SendMessage': self.process_send_message_response,
            'SendMessageBatch': self.process_send_message_batch_response,
            'ReceiveMessage': self.process_receive_message_response
        })
        _, request_data = args
        self.request_data = request_data
        self.response = response

        super(BotocoreSQSEvent, self).__init__(
            wrapped,
            instance,
            args,
            kwargs,
            start_time,
            response,
            exception
        )

        if 'QueueUrl' in request_data:
            self.resource['name'] = request_data['QueueUrl'].split('/')[-1]
        elif 'QueueName' in request_data:
            self.resource['name'] = request_data['QueueName']

        # Currently tracing only first entry
        entry = request_data['Entries'][0] if (
                'Entries' in request_data) else request_data
        if 'MessageBody' in entry:
            add_data_if_needed(
                self.resource['metadata'],
                'Message Body',
                entry['MessageBody']
            )

    def update_response(self, response):
        """
        Adds response data to event.
        :param response: Response from botocore
        """
        super(BotocoreSQSEvent, self).update_response(response)
        self.RESPONSE_TO_FUNC.get(self.resource['operation'], empty_func)()

    def process_send_message_response(self):
        """
        Process the send message response
        """
        self.resource['metadata']['Message ID'] = self.response['MessageId']
        self.resource['metadata']['MD5 Of Message Body'] = (
            self.response['MD5OfMessageBody']
        )

    def process_send_message_batch_response(self):
        """
        Process the send message batch response
        """
        # Check for at least one message sent successfully
        if not self.response.get('Successful'):
            return
        message = self.response['Successful'][0]
        self.resource['metadata']['Message ID'] = message['MessageId']
        self.resource['metadata']['MD5 Of Message Body'] = (
            message['MD5OfMessageBody']
        )

    # pylint: disable=invalid-name
    def process_receive_message_response(self):
        """
        Process the receive message response -
            notice that only first message is traced
        """
        if 'Messages' in self.response:
            messages_number = len(self.response['Messages'])
            self.resource['metadata']['Message ID'] = (
                self.response['Messages'][0]['MessageId']
            )
            self.resource['metadata']['MD5 Of Message Body'] = (
                self.response['Messages'][0]['MD5OfBody']
            )
        else:
            messages_number = 0

        self.resource['metadata']['Number Of Messages'] = messages_number


class BotocoreDynamoDBEvent(BotocoreEvent):
    """
    Represents DynamoDB botocore event.
    """
    RESOURCE_TYPE = 'dynamodb'
    CONDITION_FIELDS = ['FilterExpression', 'KeyConditionExpression']

    def __init__(self, wrapped, instance, args, kwargs, start_time, response,
                 exception):
        """
        Initialize.
        :param wrapped: wrapt's wrapped
        :param instance: wrapt's instance
        :param args: wrapt's args
        :param kwargs: wrapt's kwargs
        :param start_time: Start timestamp (epoch)
        :param response: response data
        :param exception: Exception (if happened)
        """
        self.RESPONSE_TO_FUNC.update({
            'Scan': self.process_query_and_scan_response,
            'Query': self.process_query_and_scan_response,
            'GetItem': self.process_get_item_response,
            'BatchGetItem': self.process_batch_get_response,
            'ListTables': self.process_list_tables_response
        })

        self.OPERATION_TO_FUNC.update({
            'TransactWriteItems': self.process_transact_write_op,
            'PutItem': self.process_put_item_op,
            'UpdateItem': self.process_update_item_op,
            'GetItem': self.process_get_item_op,
            'DescribeTable': self.process_describe_table_op,
            'DeleteItem': self.process_delete_item_op,
            'BatchWriteItem': self.process_batch_write_op,
            'BatchGetItem': self.process_batch_get_op,
            'Scan': self.process_scan_op,
            'Query': self.process_query_op
        })

        _, request_data = args
        self.request_data = request_data
        self.response = response
        self.deserializer = TypeDeserializer() if TypeDeserializer else None

        super(BotocoreDynamoDBEvent, self).__init__(
            wrapped,
            instance,
            args,
            kwargs,
            start_time,
            response,
            exception
        )
        self.OPERATION_TO_FUNC.get(self.resource['operation'], empty_func)()

    def update_response(self, response):
        """
        Adds response data to event.
        :param response: Response from botocore
        """
        super(BotocoreDynamoDBEvent, self).update_response(response)
        self.RESPONSE_TO_FUNC.get(self.resource['operation'], empty_func)()

    def process_get_item_op(self):
        """
        Process the get item operation.
        """
        self.resource['name'] = self.request_data['TableName']
        self.resource['metadata']['Key'] = self.request_data['Key']

    def process_put_item_op(self):
        """
        Process the put item operation.
        """
        self.resource['name'] = self.request_data['TableName']
        if 'Item' in self.request_data:
            item = self.request_data['Item']
            add_data_if_needed(self.resource['metadata'], 'Item', item)
            self.store_item_hash(item)

    def process_delete_item_op(self):
        """
        Process the delete item operation.
        """
        self.resource['name'] = self.request_data['TableName']
        deserialized_key = self._deserialize_item(self.request_data['Key'])
        if deserialized_key is not None:
            self.resource['metadata']['item_hash'] = hashlib.md5(
                json.dumps(deserialized_key).encode('utf-8')
            ).hexdigest()
        add_data_if_needed(
            self.resource['metadata'],
            'Key',
            self.request_data['Key']
        )

    def process_describe_table_op(self):
        """
        Process the describe tables operation.
        """
        self.resource['name'] = self.request_data['TableName']

    def process_transact_write_op(self):
        """
        Process the TransactWriteItems operation.
        """
        add_data_if_needed(
            self.resource['metadata'],
            'Items',
            self.request_data.get('TransactItems')
        )

    def process_update_item_op(self):
        """
        Process the update item operation.
        """
        self.resource['name'] = self.request_data['TableName']
        deserialized_key = self._deserialize_item(self.request_data['Key'])
        if deserialized_key is not None:
            self.resource['metadata']['item_hash'] = hashlib.md5(
                json.dumps(deserialized_key).encode('utf-8')
            ).hexdigest()

        self.resource['metadata']['Update Parameters'] = {
            'Key': self.request_data['Key'],
            'Expression Attribute Values': self.request_data.get(
                'ExpressionAttributeValues', None),
            'Update Expression': self.request_data.get(
                'UpdateExpression',
                None
            ),
        }

    def process_batch_write_op(self):
        """
        Process the batch write operation.
        """
        table_name = list(self.request_data['RequestItems'].keys())[0]
        self.resource['name'] = table_name
        added_items = []
        deleted_keys = []
        for item in self.request_data['RequestItems'][table_name]:
            if 'PutRequest' in item:
                added_items.append(item['PutRequest']['Item'])
            if 'DeleteRequest' in item:
                deleted_keys.append(item['DeleteRequest']['Key'])

        if deleted_keys:
            add_data_if_needed(
                self.resource['metadata'],
                'Deleted Keys',
                deleted_keys
            )
        if added_items:
            add_data_if_needed(
                self.resource['metadata'],
                'Added Items',
                added_items
            )

    def process_batch_get_op(self):
        """
        Process the batch get item operation.
        """
        table_name = list(self.request_data['RequestItems'].keys())[0]
        self.resource['name'] = table_name
        keys = list(self.request_data['RequestItems'][table_name]['Keys'])

        self.resource['metadata']['Keys'] = keys

    def process_query_op(self):
        """
        Process the query operation.
        """
        self.resource['name'] = self.request_data['TableName']
        if ConditionExpressionBuilder is None:
            return

        request_data = self._stringify_conditions(self.request_data.copy())
        if trace_factory.metadata_only:
            # Remove parameters containing non-metadata
            data_parameters = [
                'KeyConditions',
                'QueryFilter',
                'ExclusiveStartKey',
                'ProjectionExpression',
                'FilterExpression',
                'KeyConditionExpression',
                'ExpressionAttributeValues'
            ]
            for parameter in data_parameters:
                request_data.pop(parameter, None)
        self.resource['metadata']['Parameters'] = request_data

    def process_scan_op(self):
        """
        Process the scan operation.
        """
        self.resource['name'] = self.request_data['TableName']
        if ConditionExpressionBuilder is None:
            return

        request_data = self._stringify_conditions(self.request_data.copy())
        if trace_factory.metadata_only:
            # Remove parameters containing non-metadata
            data_parameters = [
                'ScanFilter'
                'ExclusiveStartKey',
                'ProjectionExpression',
                'FilterExpression',
                'ExpressionAttributeValues'
            ]
            for parameter in data_parameters:
                request_data.pop(parameter, None)
        self.resource['metadata']['Parameters'] = request_data

    def process_get_item_response(self):
        """
        Process the get item response.
        """
        self.resource['name'] = self.request_data['TableName']
        if self.response.get('Item'):
            add_data_if_needed(
                self.resource['metadata'],
                'Item',
                self.response['Item']
            )
        else:
            self.resource['metadata']['Item'] = 'null'

    def process_batch_get_response(self):
        """
        Process the batch get item response.
        """
        table_name = list(self.request_data['RequestItems'].keys())[0]
        self.resource['name'] = table_name
        add_data_if_needed(
            self.resource['metadata'],
            'Items',
            self.response['Responses'][table_name]
        )

    def process_list_tables_response(self):
        """
        Process the list tables response.
        """
        self.resource['name'] = 'DynamoDBEngine'
        self.resource['metadata']['Table Names'] = \
            ', '.join(self.response['TableNames'])

    def process_query_and_scan_response(self):
        """
        Process the query/scan response
        """
        response_data = self.response.copy()
        if trace_factory.metadata_only:
            # Remove parameters containing non-metadata
            response_data.pop('Items', None)
            response_data.pop('LastEvaluatedKey', None)
        self.resource['metadata']['Response'] = response_data

    def store_item_hash(self, item):
        """
        Store the item hash in the metadata.
        :param item: The item to store the hash for.
        """
        deserializer = TypeDeserializer()

        # Try to deserialize the data in order to remove dynamoDB data types.
        for key in item:
            try:
                item[key] = deserializer.deserialize(item[key])
            except (TypeError, AttributeError):
                break
        self.resource['metadata']['item_hash'] = hashlib.md5(
            json.dumps(item, sort_keys=True).encode('utf-8')).hexdigest()

    def _deserialize_item(self, item):
        """
        Deserialize DynamoDB Item in order to remove types definitions.
        :param item: The item to deserialize.
        :return: Deserialized item.
        """
        if self.deserializer is None:
            return None
        deserialized_item = item.copy()
        for key in item:
            try:
                deserialized_item[key] = self.deserializer.deserialize(
                    item[key]
                )
            except (TypeError, AttributeError):
                break
        return deserialized_item

    def _stringify_conditions(self, request_data):
        """
        convert Key and Attr object of DynamoDB to strings.
        :param request_data: DynamoDB request data
        :return: updated request data
        """
        for field in self.CONDITION_FIELDS:
            if field in request_data:
                try:
                    request_data[field] = str(
                        ConditionExpressionBuilder().build_expression(
                            request_data[field],
                            field == 'KeyConditionExpression'
                        )
                    )
                except Exception:  # pylint: disable=W0703
                    pass
        return request_data


class BotocoreSESEvent(BotocoreEvent):
    """
    Represents SES botocore event.
    """
    RESOURCE_TYPE = 'ses'

    def __init__(self, wrapped, instance, args, kwargs, start_time, response,
                 exception):
        """
        Initialize.
        :param wrapped: wrapt's wrapped
        :param instance: wrapt's instance
        :param args: wrapt's args
        :param kwargs: wrapt's kwargs
        :param start_time: Start timestamp (epoch)
        :param response: response data
        :param exception: Exception (if happened)
        """

        super(BotocoreSESEvent, self).__init__(
            wrapped,
            instance,
            args,
            kwargs,
            start_time,
            response,
            exception
        )

        _, request_data = args

        if self.resource['operation'] == 'SendEmail':
            self.resource['metadata']['source'] = request_data['Source']
            self.resource['metadata']['destination'] = \
                request_data['Destination']
            self.resource['metadata']['subject'] = \
                request_data['Message']['Subject']

            add_data_if_needed(
                self.resource['metadata'],
                'body',
                request_data['Message']['Body']
            )

    def update_response(self, response):
        """
        Adds response data to event.
        :param response: Response from botocore
        :return: None
        """

        super(BotocoreSESEvent, self).update_response(response)

        if self.resource['operation'] == 'SendEmail':
            self.resource['metadata']['message_id'] = response['MessageId']


class BotocoreAthenaEvent(BotocoreEvent):
    """
    Represents Athena botocore event
    """
    RESOURCE_TYPE = 'athena'

    def __init__(self, wrapped, instance, args, kwargs, start_time, response,
                 exception):
        self.RESPONSE_TO_FUNC.update(
            {'GetQueryExecution': self.process_get_query_response,
             'GetQueryResults': self.process_query_results_response,
             'StartQueryExecution': self.process_start_query_response,
             }
        )

        self.OPERATION_TO_FUNC.update(
            {'StartQueryExecution': self.process_start_query_operation,
             'GetQueryExecution': self.process_general_query_operation,
             'GetQueryResults': self.process_general_query_operation,
             'StopQueryExecution': self.process_general_query_operation,
             }
        )

        super(BotocoreAthenaEvent, self).__init__(
            wrapped,
            instance,
            args,
            kwargs,
            start_time,
            response,
            exception
        )

        self.OPERATION_TO_FUNC.get(
            self.resource['operation'], empty_func)(args, kwargs)

    def update_response(self, response):
        """
        Adds response data to event.
        :param response: Response from botocore's Athena Client
        """
        super(BotocoreAthenaEvent, self).update_response(response)

        self.RESPONSE_TO_FUNC.get(
            self.resource['operation'], empty_func)(response)

    def process_start_query_operation(self, args, _):
        """
        Process StartQueryExecution operation
        :param args: command arguments
        :param _: unused, kwargs
        :return: None
        """
        _, request_args = args

        self.resource['metadata']['Database'] = \
            request_args.get('QueryExecutionContext', {}).get('Database', None)

        add_data_if_needed(self.resource['metadata'], 'Query',
                           request_args['QueryString'])

    def process_get_query_response(self, response):
        """
        Process GetQueryExecution response
        :param response: response from Athena Client
        :return: None
        """
        metadata = self.resource['metadata']

        response_details = response['QueryExecution']

        metadata['Query Status'] = \
            response_details.get('Status', {}).get('State', None)

        metadata['Result Location'] = \
            response_details.get('ResultConfiguration', {}). \
                get('OutputLocation', None)

        metadata['Query ID'] = response_details['QueryExecutionId']
        add_data_if_needed(metadata, 'Query', response_details['Query'])

    def process_general_query_operation(self, args, _):
        """
        Process generic Athena Query operations
        :param args: command arguments
        :param _: unused, kwargs
        :return: None
        """
        _, request_args = args
        self.resource['metadata']['Query ID'] = request_args['QueryExecutionId']

    def process_query_results_response(self, response):
        """
        Process GetQueryResults response
        :param response: response from Athena Client
        :return: None
        """
        self.resource['metadata']['Query Rows Count'] = \
            len(response.get('ResultSet', {}).get('Rows', []))

    def process_start_query_response(self, response):
        """
        Process StartQueryExecution response
        :param response: response from Athena Client
        :return: None
        """
        self.resource['metadata']['Query ID'] = response['QueryExecutionId']


class BotocoreFirehoseEvent(BotocoreEvent):
    """
    Represents Firehose botocore event
    """
    RESOURCE_TYPE = 'firehose'

    def __init__(self, wrapped, instance, args, kwargs, start_time, response,
                 exception):
        self.RESPONSE_TO_FUNC.update({
            'PutRecord': self.put_record_response,
            'PutRecordBatch': self.put_record_batch_response,
        })

        self.OPERATION_TO_FUNC.update({
            'PutRecord': self.put_record_operation,
            'PutRecordBatch': self.put_record_batch_operation,
        })

        super(BotocoreFirehoseEvent, self).__init__(
            wrapped,
            instance,
            args,
            kwargs,
            start_time,
            response,
            exception
        )

        self.OPERATION_TO_FUNC.get(self.resource['operation'], empty_func)(
            args,
            kwargs
        )

    def update_response(self, response):
        """
        Adds response data to event.
        :param response: Response from botocore's Firehose Client
        """
        super(BotocoreFirehoseEvent, self).update_response(response)

        self.RESPONSE_TO_FUNC.get(self.resource['operation'], empty_func)(
            response
        )

    def put_record_operation(self, args, _):
        """
        Process PutRecord operation
        :param args: command arguments
        :param _: unused, kwargs
        :return: None
        """
        _, request_args = args
        self.resource['name'] = request_args['DeliveryStreamName']

    def put_record_response(self, response):
        """
        Process PutRecord response
        :param response: response from Firehose Client
        :return: None
        """
        self.resource['metadata']['record_id'] = response['RecordId']

    def put_record_batch_operation(self, args, _):
        """
        Process PutRecordBatch operation
        :param args: command arguments
        :param _: unused, kwargs
        :return: None
        """
        _, request_args = args
        self.resource['name'] = request_args['DeliveryStreamName']
        self.resource['metadata']['records_count'] = len(
            request_args['Records']
        )

    def put_record_batch_response(self, response):
        """
        Process PutRecordBatch response
        :param response: response from Firehose Client
        :return: None
        """
        metadata = self.resource['metadata']
        metadata['Failed Put Count'] = response['FailedPutCount']
        if response['RequestResponses']:
            metadata['record_id'] = response['RequestResponses'][0]['RecordId']


class BotocoreCognitoEvent(BotocoreEvent):
    """
    Represents Cognito botocore event
    """
    RESOURCE_TYPE = 'cognitoidentityprovider'
    RESOURCE_TYPE_UPDATE = 'cognito-idp'


    def __init__(self, wrapped, instance, args, kwargs, start_time, response,
                 exception):
        self.RESPONSE_TO_FUNC.update({
            'AdminCreateUser': self.admin_create_user_res,
            'AdminListGroupsForUser': self.admin_list_user_group_res,
        })

        self.OPERATION_TO_FUNC.update({
            'AdminCreateUser': self.general_user_pool_op,
            'AdminInitiateAuth': self.general_user_pool_op,
            'AdminListGroupsForUser': self.general_user_pool_op,
            'AdminSetUserPassword': self.admin_set_pass_op,
            'AdminRespondToAuthChallenge': self.general_user_pool_op,
            'DescribeUserPool': self.general_user_pool_op,
            'ListUsers': self.general_user_pool_op,
            'UpdateUserPool': self.general_user_pool_op,
            'SignUp': self.general_user_pool_client_op,
        })

        super(BotocoreCognitoEvent, self).__init__(
            wrapped,
            instance,
            args,
            kwargs,
            start_time,
            response,
            exception
        )
        self.resource['type'] = self.RESOURCE_TYPE_UPDATE

        self.OPERATION_TO_FUNC.get(self.resource['operation'], empty_func)(
            args,
            kwargs
        )

    def update_response(self, response):
        """
        Adds response data to event.
        :param response: Response from botocore's Cognito Client
        """
        super(BotocoreCognitoEvent, self).update_response(response)

        self.RESPONSE_TO_FUNC.get(self.resource['operation'], empty_func)(
            response
        )

    def admin_create_user_res(self, response):
        """
        Process AdminCreateUser response
        :param response: response from Client
        :return: None
        """
        self.resource['metadata']['user'] = response['User']

    def admin_list_user_group_res(self, response):
        """
        Process AdminListGroupsForUser response
        :param response: response from Client
        :return: None
        """
        add_data_if_needed(
            self.resource['metadata'],
            'groups',
            response['Groups']
        )

    def admin_set_pass_op(self, args, _):
        """
        Process AdminSetUserPassword operation
        :param args: command arguments
        :param _: unused, kwargs
        :return: None
        """
        _, request_args = args
        self.resource['name'] = request_args['UserPoolId']
        self.resource['metadata']['username'] = request_args['Username']
        self.resource['metadata']['permanent'] = request_args.get(
            'Permanent',
            False
        )

    def general_user_pool_op(self, args, _):
        """
        Process any User Pool operation
        :param args: command arguments
        :param _: unused, kwargs
        :return: None
        """
        _, request_args = args
        self.resource['name'] = request_args.get('UserPoolId')
        self.resource['metadata'].update(request_args)

    def general_user_pool_client_op(self, args, _):
        """
        Process any User Pool App Client operation
        :param args: command arguments
        :param _: unused, kwargs
        :return: None
        """
        _, request_args = args
        self.resource['name'] = request_args.get('ClientId')
        self.resource['metadata'].update(request_args)


class BotocoreKMSEvent(BotocoreEvent):
    """
    Represents KMS botocore event
    """
    RESOURCE_TYPE = 'kms'

    def __init__(self, wrapped, instance, args, kwargs, start_time, response,
                 exception):
        self.RESPONSE_TO_FUNC.update({
            'Decrypt': self.decrypt_response,
            'Encrypt': self.encrypt_response,
        })

        self.OPERATION_TO_FUNC.update({
            'Decrypt': self.decrypt_operation,
            'Encrypt': self.encrypt_operation,
        })

        super(BotocoreKMSEvent, self).__init__(
            wrapped,
            instance,
            args,
            kwargs,
            start_time,
            response,
            exception
        )

        self.resource['name'] = 'KMS'
        self.OPERATION_TO_FUNC.get(self.resource['operation'], empty_func)(
            args,
            kwargs
        )

    def update_response(self, response):
        """
        Adds response data to event.
        :param response: Response from botocore's KMS Client
        """
        super(BotocoreKMSEvent, self).update_response(response)

        self.RESPONSE_TO_FUNC.get(self.resource['operation'], empty_func)(
            response
        )

    def decrypt_operation(self, args, _):
        """
        Process Decrypt operation
        :param args: command arguments
        :param _: unused, kwargs
        :return: None
        """
        _, request_args = args
        self.resource['metadata']['cipher_blob_size'] = len(
            request_args['CiphertextBlob']
        )

    def decrypt_response(self, response):
        """
        Process Decrypt response
        :param response: response from KMS Client
        :return: None
        """
        self.resource['metadata']['key_id'] = response['KeyId']
        self.resource['metadata']['plaintext_size'] = len(response['Plaintext'])

    def encrypt_operation(self, args, _):
        """
        Process Encrypt operation
        :param args: command arguments
        :param _: unused, kwargs
        :return: None
        """
        _, request_args = args
        self.resource['metadata']['key_id'] = request_args['KeyId']
        self.resource['metadata']['plaintext_size'] = len(
            request_args['Plaintext']
        )

    def encrypt_response(self, response):
        """
        Process Encrypt response
        :param response: response from KMS Client
        :return: None
        """
        self.resource['metadata']['cipher_blob_size'] = len(
            response['CiphertextBlob']
        )


class BotocoreSSMEvent(BotocoreEvent):
    """
    Represents SSM botocore event
    """
    RESOURCE_TYPE = 'ssm'

    def __init__(self, wrapped, instance, args, kwargs, start_time, response,
                 exception):
        self.RESPONSE_TO_FUNC.update({
            'GetParameters': self.get_params_response,
        })

        self.OPERATION_TO_FUNC.update({
            'GetParameters': self.get_params_operation,
        })

        super(BotocoreSSMEvent, self).__init__(
            wrapped,
            instance,
            args,
            kwargs,
            start_time,
            response,
            exception
        )

        self.resource['name'] = 'SSM'
        self.OPERATION_TO_FUNC.get(self.resource['operation'], empty_func)(
            args,
            kwargs
        )

    def update_response(self, response):
        """
        Adds response data to event.
        :param response: Response from botocore's SSM Client
        """
        super(BotocoreSSMEvent, self).update_response(response)

        self.RESPONSE_TO_FUNC.get(self.resource['operation'], empty_func)(
            response
        )

    def get_params_operation(self, args, _):
        """
        Process GetParameters operation
        :param args: command arguments
        :param _: unused, kwargs
        :return: None
        """
        _, request_args = args
        add_data_if_needed(
            self.resource['metadata'],
            'Names',
            str(request_args.get('Names', '')),
        )
        self.resource['metadata']['With Decryption'] = (
            request_args.get('WithDecryption', True)
        )

    def get_params_response(self, response):
        """
        Process GetParameters response
        :param response: response from SSM Client
        :return: None
        """
        add_data_if_needed(
            self.resource['metadata'],
            'Parameters',
            response['Parameters']
        )
        add_data_if_needed(
            self.resource['metadata'],
            'Invalid Parameters',
            str(response.get('InvalidParameters', ''))
        )


class BotocoreStepFunctionEvent(BotocoreEvent):
    """
    Represents Step Function botocore event
    """
    RESOURCE_TYPE = 'sfn'
    REAL_RESOURCE_TYPE = 'stepfunctions'
    DEFAULT_EXECTUTION_NAME = 'Unnamed Execution'

    def __init__(self, wrapped, instance, args, kwargs, start_time, response,
                 exception):
        self.RESPONSE_TO_FUNC.update({
            'StartExecution': self.process_start_exec_response,
            'SendTaskHeartbeat': self.process_send_task_heartbeat_response,
            'DescribeExecution': self.process_describe_execution_response,
        })

        self.OPERATION_TO_FUNC.update({
            'StartExecution': self.process_start_exec_operation,
            'SendTaskSuccess': self.process_send_task_success_exec_operation,
            'SendTaskHeartbeat': self.process_send_task_heartbeat_operation,
            'DescribeExecution': self.process_describe_execution_operation,
        })

        super(BotocoreStepFunctionEvent, self).__init__(
            wrapped,
            instance,
            args,
            kwargs,
            start_time,
            response,
            exception
        )

        # Fix resource type
        self.resource['type'] = self.REAL_RESOURCE_TYPE
        self.OPERATION_TO_FUNC.get(
            self.resource['operation'],
            empty_func
        )(args, kwargs)

    def initialize_step_dict(self, request_args, params_property_name):
        try:
            machine_input = json.loads(request_args[params_property_name])
            self.resource['metadata']['steps_dict'] = (
                machine_input[STEP_DICT_NAME]
            )
        except Exception:  # pylint: disable=broad-except
            pass

    def update_response(self, response):
        """
        Adds response data to event.
        :param response: Response from botocore's Athena Client
        """
        super(BotocoreStepFunctionEvent, self).update_response(response)

        self.RESPONSE_TO_FUNC.get(
            self.resource['operation'],
            empty_func
        )(response)

    def process_start_exec_operation(self, args, _):
        """
        Process startExecution operation
        :param args: command arguments
        :param _: unused, kwargs
        :return: None
        """
        _, request_args = args
        self.resource['name'] = self.REAL_RESOURCE_TYPE
        self.resource['metadata']['State Machine ARN'] = (
            request_args['stateMachineArn']
        )
        if 'name' in request_args:
            self.resource['metadata']['Execution Name'] = request_args['name']

        add_data_if_needed(
            self.resource['metadata'],
            'Input',
            request_args['input']
        )

        self.initialize_step_dict(request_args, 'input')

    def process_send_task_success_exec_operation(self, args, _):
        """
        Process sendTaskSuccess operation
        :param args: command arguments
        :param _: unused, kwargs
        :return: None
        """
        _, request_args = args

        self.initialize_step_dict(request_args, 'output')

    def process_send_task_heartbeat_operation(self, args, _):
        """
        Process sendTaskSuccess operation
        :param args: command arguments
        :param _: unused, kwargs
        :return: None
        """
        _, request_args = args
        add_metadata_from_dict(self.resource, request_args, 'taskToken')

    def process_describe_execution_operation(self, args, _):
        """
        Process sendTaskSuccess operation
        :param args: command arguments
        :param _: unused, kwargs
        :return: None
        """
        _, request_args = args
        add_metadata_from_dict(self.resource, request_args,
                                      'executionArn')

    def process_start_exec_response(self, response):
        """
        Process startExecution response
        :param response: response from Step function Client
        :return: None
        """
        add_metadata_from_dict(self.resource, response, 'executionArn')

    def process_send_task_heartbeat_response(self, response):
        """
        Process sendTaskHeartbeat response
        :param response: response from Step function Client
        :return: None
        """

        add_metadata_from_dict(self.resource, response,
                                      'ResponseMetadata')

    def process_describe_execution_response(self, response):
        """
        Process describeExecution response
        :param response: response from Step function Client
        :return: None
        """
        if isinstance(response, dict):
            for key in response:
                add_metadata_from_dict(self.resource, response, key)


class BotocoreLambdaEvent(BotocoreEvent):
    """
    Represents lambda botocore event.
    """

    RESOURCE_TYPE = 'lambda'
    AWS_ACCOUNT_IND = 4

    def __init__(self, wrapped, instance, args, kwargs, start_time, response,
                 exception):
        """
        Initialize.
        :param wrapped: wrapt's wrapped
        :param instance: wrapt's instance
        :param args: wrapt's args
        :param kwargs: wrapt's kwargs
        :param start_time: Start timestamp (epoch)
        :param response: response data
        :param exception: Exception (if happened)
        """
        super(BotocoreLambdaEvent, self).__init__(
            wrapped,
            instance,
            args,
            kwargs,
            start_time,
            response,
            exception
        )

        _, request_data = args
        func_name = request_data.get('FunctionName', '')
        if ':' in func_name:
            splitted_func_name = func_name.split(':')
            self.resource['metadata']['aws_account'] = splitted_func_name[
                self.AWS_ACCOUNT_IND
            ]
            func_name = splitted_func_name[-1]
        self.resource['name'] = func_name
        if 'Payload' in request_data:
            add_data_if_needed(
                self.resource['metadata'],
                'payload',
                request_data['Payload']
            )
        if 'InvokeArgs' in request_data and \
                isinstance(request_data['InvokeArgs'], str):
            add_data_if_needed(
                self.resource['metadata'],
                'payload',
                request_data['InvokeArgs']
            )


class BotocoreEmr(BotocoreEvent):
    """
    Represents EMR botocore event.
    """

    RESOURCE_TYPE = 'emr'

    def __init__(self, wrapped, instance, args, kwargs, start_time, response,
                 exception):
        """
        Initialize.
        :param wrapped: wrapt's wrapped
        :param instance: wrapt's instance
        :param args: wrapt's args
        :param kwargs: wrapt's kwargs
        :param start_time: Start timestamp (epoch)
        :param response: response data
        :param exception: Exception (if happened)
        """

        self.OPERATION_TO_FUNC.update({
            'AddJobFlowSteps': self.add_job_flow_steps_op,
            'TerminateJobFlows': self.terminate_job_flows_op,
            'ListClusters': self.list_clusters_op,
            'RunJobFlow': self.run_job_flow_op
        })

        self.RESPONSE_TO_FUNC.update({
            'DescribeCluster': self.describe_cluster_response,
            'DescribeStep': self.describe_step_response,
            'AddJobFlowSteps': self.add_job_flow_steps_response,
            'ListInstances': self.list_instances_response,
            'ListClusters': self.list_clusters_response,
            'RunJobFlow': self.run_job_flow_response
        })
        _, request_data = args
        self.request_data = request_data
        self.response = response

        super(BotocoreEmr, self).__init__(
            wrapped,
            instance,
            args,
            kwargs,
            start_time,
            response,
            exception
        )

        self.resource['name'] = request_data.get(
            'ClusterId',
            request_data.get('JobFlowId', 'EMR')
        )
        self.OPERATION_TO_FUNC.get(self.resource['operation'], empty_func)()

    def add_job_flow_steps_op(self):
        """
        Handle add job flow steps operation.
        """
        self.resource['metadata']['Steps'] = self.request_data['Steps']

    def terminate_job_flows_op(self):
        """
        Handle terminate job flows operation.
        """
        self.resource['metadata']['Job Flow IDs'] = self.request_data[
            'JobFlowIds']

    def list_clusters_op(self):
        """
        Handle list clusters operation.
        """
        self.resource['metadata']['Cluster States'] = self.request_data.get(
            'ClusterStates')

    def run_job_flow_op(self):
        """
        Handle run job flow operation.
        """
        self.resource['metadata']['Request'] = self.request_data

    def update_response(self, response):
        """
        Adds response data to event.
        :param response: Response from botocore
        """
        super(BotocoreEmr, self).update_response(response)
        self.RESPONSE_TO_FUNC.get(self.resource['operation'], empty_func)()

    def describe_cluster_response(self):
        """
        Handle describe cluster response.
        """
        cluster = self.response['Cluster']
        self.resource['metadata']['Status'] = cluster['Status']
        self.resource['metadata']['Cluster Name'] = cluster['Name']
        self.resource['metadata']['Cluster ID'] = cluster['Id']

    def describe_step_response(self):
        """
        Handle describe step response.
        """
        step = self.response['Step']
        self.resource['metadata']['Step Id'] = step['Id']
        self.resource['metadata']['Step Name'] = step['Name']
        self.resource['metadata']['Status'] = step['Status']
        self.resource['metadata']['Step Config'] = step['Config']

    def add_job_flow_steps_response(self):
        """
        Handle add job flow steps response.
        """
        self.resource['metadata']['Step Ids'] = self.response['StepIds']

    def list_instances_response(self):
        """
        Handle list instances response.
        """
        self.resource['metadata']['Instances'] = self.response['Instances']

    def list_clusters_response(self):
        """
        Handle list clusters response.
        """
        self.resource['metadata']['Clusters'] = self.response['Clusters']

    def run_job_flow_response(self):
        """
        Handle run job flow response.
        """
        self.resource['metadata']['Job Flow ID'] = self.response['JobFlowId']


class BotocoreEventFactory(object):
    """
    Factory class, generates botocore event.
    """

    FACTORY = {
        class_obj.RESOURCE_TYPE: class_obj
        for class_obj in BotocoreEvent.__subclasses__()
    }

    @staticmethod
    def create_event(wrapped, instance, args, kwargs, start_time, response,
                     exception):
        """
        Create an event according to the given instance_type.
        :param wrapped:
        :param instance:
        :param args:
        :param kwargs:
        :param start_time:
        :param response:
        :param exception:
        :return:
        """
        instance_type = instance.__class__.__name__.lower()
        event_class = BotocoreEventFactory.FACTORY.get(
            instance_type,
            None
        )

        if event_class is not None:
            event = event_class(
                wrapped,
                instance,
                args,
                kwargs,
                start_time,
                response,
                exception
            )
            trace_factory.add_event(event)



"""
Tornado patcher module.
"""

from __future__ import absolute_import
import time
import traceback
import uuid
from functools import partial
import wrapt
from tornado.httpclient import HTTPRequest
from tornado.httputil import HTTPHeaders
import epsagon.trace
from epsagon.modules.general_wrapper import wrapper
from epsagon.runners.tornado import TornadoRunner
from epsagon.http_filters import ignore_request, is_ignored_endpoint
from epsagon.utils import (
    collect_container_metadata,
    print_debug,
    get_epsagon_http_trace_id
)
from ..constants import EPSAGON_HEADER
from ..events.tornado_client import TornadoClientEventFactory


TORNADO_TRACE_ID = 'epsagon_tornado_trace_key'


class TornadoWrapper(object):
    """
    Wraps Tornado web framework to get requests.
    """
    RUNNERS = {}

    @classmethod
    def before_request(cls, wrapped, instance, args, kwargs):
        """
        Runs when new request comes in.
        :param wrapped: wrapt's wrapped
        :param instance: wrapt's instance
        :param args: wrapt's args
        :param kwargs: wrapt's kwargs
        """
        print_debug('before_request Tornado request')
        try:
            ignored = ignore_request('', instance.request.path)
            if not ignored and not is_ignored_endpoint(instance.request.path):
                unique_id = str(uuid.uuid4())
                trace = epsagon.trace.trace_factory.get_or_create_trace(
                    unique_id=unique_id
                )

                trace.prepare()

                setattr(instance, TORNADO_TRACE_ID, unique_id)

                cls.RUNNERS[unique_id] = (
                    TornadoRunner(time.time(), instance.request)
                )

                trace.set_runner(cls.RUNNERS[unique_id])
                print_debug('Created Tornado Runner')

                # Collect metadata in case this is a container.
                collect_container_metadata(
                    cls.RUNNERS[unique_id].resource['metadata']
                )
        except Exception as instrumentation_exception:  # pylint: disable=W0703
            epsagon.trace.trace_factory.add_exception(
                instrumentation_exception,
                traceback.format_exc()
            )
        return wrapped(*args, **kwargs)

    @classmethod
    def after_request(cls, wrapped, instance, args, kwargs):
        """
        Runs after first process of response.
        :param wrapped: wrapt's wrapped
        :param instance: wrapt's instance
        :param args: wrapt's args
        :param kwargs: wrapt's kwargs
        """
        print_debug('after_request Tornado request')
        response_body = None
        try:
            response_body = getattr(instance, '_write_buffer', None)
            if response_body and isinstance(response_body, list):
                response_body = b''.join(response_body)
        except Exception as instrumentation_exception:  # pylint: disable=W0703
            epsagon.trace.trace_factory.add_exception(
                instrumentation_exception,
                traceback.format_exc()
            )

        res = wrapped(*args, **kwargs)
        trace = None
        is_trace_sent = False
        try:
            unique_id = getattr(instance, TORNADO_TRACE_ID, None)
            if not unique_id:
                return res

            tornado_runner = cls.RUNNERS.pop(unique_id)

            trace = epsagon.trace.trace_factory.switch_active_trace(
                unique_id
            )

            # Ignoring 404s
            if getattr(instance, '_status_code', None) == 404:
                epsagon.trace.trace_factory.pop_trace(trace=trace)
                print_debug('Ignoring 404 Tornado request')
                return res

            if trace:
                content = (
                    instance._headers.get(  # pylint: disable=protected-access
                        'Content-Type',
                        ''
                    )
                )
                ignored = ignore_request(content, '')
                if not ignored:
                    tornado_runner.update_response(instance, response_body)
                    epsagon.trace.trace_factory.send_traces(trace)
                    is_trace_sent = True
        except Exception:  # pylint: disable=W0703
            if not is_trace_sent and trace:
                epsagon.trace.trace_factory.pop_trace(trace=trace)

        return res

    @classmethod
    def collect_exception(cls, wrapped, instance, args, kwargs):
        """
        Runs after first process of response.
        :param wrapped: wrapt's wrapped
        :param _: wrapt's instance
        :param args: wrapt's args
        :param kwargs: wrapt's kwargs
        """
        print_debug('collect_exception Tornado request')
        try:
            unique_id = getattr(instance, TORNADO_TRACE_ID, None)

            if unique_id and cls.RUNNERS.get(unique_id):
                _, exception, _ = args
                cls.RUNNERS[unique_id].set_exception(
                    exception, traceback.format_exc()
                )
        except Exception as instrumentation_exception:  # pylint: disable=W0703
            epsagon.trace.trace_factory.add_exception(
                instrumentation_exception,
                traceback.format_exc()
            )

        return wrapped(*args, **kwargs)

    @classmethod
    def run_callback(cls, wrapped, _, args, kwargs):
        """
        Instrument run_callback in the event loop
        :param wrapped: wrapt's wrapped
        :param _: wrapt's instnace
        :param args: wrapper arguments containing the callback function
        :param kwargs: kwargs arguments
        :return: callback wrapped after selecting active trace
        """
        try:
            func = args[0]
            if isinstance(func, partial):
                func = func.func
            unique_id = getattr(func, TORNADO_TRACE_ID, None)
            if unique_id:
                epsagon.trace.trace_factory.switch_active_trace(unique_id)
        except Exception as instrumentation_exception:  # pylint: disable=W0703
            epsagon.trace.trace_factory.add_exception(
                instrumentation_exception,
                traceback.format_exc()
            )

        return wrapped(*args, **kwargs)

    @classmethod
    def wrap(cls, wrapped, _, args, kwargs):
        """
        Instrument stack context's wrap function
        :param wrapped:
        :param _:
        :param args:
        :param kwargs:
        :return:
        """
        res = wrapped(*args, **kwargs)
        if res and not hasattr(res, TORNADO_TRACE_ID):
            trace = epsagon.trace.trace_factory.active_trace
            if trace and trace.unique_id:
                setattr(res, TORNADO_TRACE_ID, trace.unique_id)
        return res

    @classmethod
    def thread_pool_submit(cls, func, _, args, kwargs):
        """
        Submits a new worker to the thread pool, wrapped in our injector
        """
        unique_id = None

        try:
            trace = epsagon.trace.trace_factory.get_trace()
            if trace and trace.unique_id in cls.RUNNERS:
                unique_id = trace.unique_id

        except Exception as instrumentation_exception:  # pylint: disable=W0703
            epsagon.trace.trace_factory.add_exception(
                instrumentation_exception,
                traceback.format_exc()
            )

        fn = args[0]
        fn_args = args[1:]
        return func(cls.thread_pool_execution, unique_id, fn, fn_args, kwargs)

    @classmethod
    def thread_pool_execution(cls, unique_id, fn, args, kwargs):
        """
            Middleware to inject unique id to the thread pool execution
        """
        try:
            if unique_id is not None:
                epsagon.trace.trace_factory.set_thread_local_unique_id(
                    unique_id
                )
        except Exception as instrumentation_exception:  # pylint: disable=W0703
            epsagon.trace.trace_factory.add_exception(
                instrumentation_exception,
                traceback.format_exc()
            )
        res = fn(*args, **kwargs)
        epsagon.trace.trace_factory.unset_thread_local_unique_id()
        return res


def _prepare_http_request(request, raise_error=True, **kwargs):
    """
    Copying parameters from original `AsyncHTTPClient.fetch` function
    :return: HTTPRequest, raise_error bool
    """
    # request can be a URL string or a HTTPRequest object
    if not isinstance(request, HTTPRequest):
        request = HTTPRequest(request, **kwargs)

    return request, raise_error


def _wrapper(wrapped, instance, args, kwargs):
    """
    General wrapper for AsyncHTTPClient instrumentation.
    :param wrapped: wrapt's wrapped
    :param instance: wrapt's instance
    :param args: wrapt's args
    :param kwargs: wrapt's kwargs
    :return: None
    """
    print_debug('AsyncHTTPClient init')
    try:
        request, raise_error = _prepare_http_request(*args, **kwargs)
    except Exception:  # pylint: disable=W0703
        return wrapped(*args, **kwargs)

    print_debug('AsyncHTTPClient setting header')
    trace_header = get_epsagon_http_trace_id()

    if isinstance(request.headers, HTTPHeaders):
        if not request.headers.get(EPSAGON_HEADER):
            request.headers.add(EPSAGON_HEADER, trace_header)
    elif isinstance(request.headers, dict):
        if EPSAGON_HEADER not in request.headers:
            request.headers[EPSAGON_HEADER] = trace_header

    print_debug('AsyncHTTPClient running wrapper')
    return wrapper(
        TornadoClientEventFactory,
        wrapped,
        instance,
        (request, ),  # new args
        {'raise_error': raise_error}  # new kwargs
    )


def patch():
    """
    Patch module.
    """
    try:
        wrapt.wrap_function_wrapper(
            'tornado.web',
            'RequestHandler._execute',
            TornadoWrapper.before_request
        )
        wrapt.wrap_function_wrapper(
            'tornado.web',
            'RequestHandler.finish',
            TornadoWrapper.after_request
        )
        wrapt.wrap_function_wrapper(
            'tornado.web',
            'RequestHandler.log_exception',
            TornadoWrapper.collect_exception
        )
        wrapt.wrap_function_wrapper(
            'tornado.ioloop',
            'IOLoop._run_callback',
            TornadoWrapper.run_callback
        )
        wrapt.wrap_function_wrapper(
            'concurrent.futures',
            'ThreadPoolExecutor.submit',
            TornadoWrapper.thread_pool_submit
        )
        wrapt.wrap_function_wrapper(
            'tornado.stack_context',
            'wrap',
            TornadoWrapper.wrap
        )
    except Exception:  # pylint: disable=broad-except
        # Can happen in different Tornado versions.
        pass

    wrapt.wrap_function_wrapper(
        'tornado.httpclient',
        'AsyncHTTPClient.fetch',
        _wrapper
    )



"""
Test Django wrapper functionality
"""

import pytest
import mock
from epsagon.wrappers.django import DjangoRequestMiddleware


TEST_BODY = 'test_body'
TEST_METHOD = 'test_method'
TEST_PATH = '/test_path'


@pytest.fixture
def test_request():
    """ A test request """
    class TestRequest:
        def __init__(self, path, method, body):
            self.path = path
            self.method = method
            self.body = body

        def get_host(self):
            return "test"

    return TestRequest(
        TEST_PATH,
        TEST_METHOD,
        TEST_BODY
    )


@mock.patch('epsagon.triggers.http.HTTPTriggerFactory')
@mock.patch('epsagon.runners.django.DjangoRunner')
@mock.patch('time.time', return_value=1)
def test_before_request(_, runner_mock, trigger_mock, test_request):
    """
    Test before_request generates a runner and a trigger.
    """
    request_mw = DjangoRequestMiddleware(test_request)

    request_mw.before_request()
    trigger_mock.factory.assert_called_once()

    runner_mock.assert_called_with(1, test_request)


@mock.patch('epsagon.triggers.http.HTTPTriggerFactory')
@mock.patch('epsagon.runners.django.DjangoRunner.update_response')
@mock.patch('time.time', return_value=1)
def test_after_request(
    _, runner_mock, trigger_mock, test_request, trace_transport
):
    """
    Test the whole  flow - a trace is generated with  the
    right content.
    """
    request_mw = DjangoRequestMiddleware(test_request)

    request_mw.before_request()
    request_mw.after_request({"content": "bla"})

    runner_mock.assert_called_once()
    trigger_mock.factory.assert_called_once()

    assert trace_transport.last_trace.events[
        0].resource['metadata']['Request Data'] == TEST_BODY
    assert trace_transport.last_trace.events[
        0].resource['metadata']['Path'] == TEST_PATH


@mock.patch('epsagon.triggers.http.HTTPTriggerFactory')
@mock.patch('epsagon.runners.django.DjangoRunner')
@mock.patch('time.time', return_value=1)
def test_ignored_path(_, runner_mock, trigger_mock, test_request):
    """
    Make sure we don't capture ignored pathes
    """
    ignored_request = test_request
    # JS files should be ignored
    ignored_request.path = "*.js"
    request_mw = DjangoRequestMiddleware(ignored_request)

    request_mw.before_request()
    trigger_mock.factory.assert_not_called()



"""
Runner for a Django Python function
"""

from __future__ import absolute_import
import uuid
from ..event import BaseEvent
from ..utils import add_data_if_needed
from ..constants import EPSAGON_HEADER_TITLE


class DjangoRunner(BaseEvent):
    """
    Represents Python Django event runner.
    """

    ORIGIN = 'runner'
    RESOURCE_TYPE = 'python_django'
    OPERATION = 'request'

    def __init__(self, start_time, request):
        """
        Initialize.
        :param start_time: event's start time (epoch).
        :param request: the incoming request.
        """
        super(DjangoRunner, self).__init__(start_time)

        self.event_id = str(uuid.uuid4())
        self.resource['name'] = request.get_host()
        self.resource['operation'] = request.method

        self.resource['metadata'].update({'Path': request.path})

        if request.body:
            add_data_if_needed(
                self.resource['metadata'],
                'Request Data',
                request.body
            )

        # request.headers introduced since django==2.2
        if hasattr(request, 'headers'):
            if request.headers.get(EPSAGON_HEADER_TITLE):
                self.resource['metadata']['http_trace_id'] = (
                    request.headers.get(EPSAGON_HEADER_TITLE)
                )

            if request.headers:
                add_data_if_needed(
                    self.resource['metadata'],
                    'Request Headers',
                    request.headers
                )

    def update_response(self, response):
        """
        Adds response data to event.
        :param response: WSGI Response
        :return: None
        """
        add_data_if_needed(
            self.resource['metadata'],
            'Response Data',
            response.content
        )

        add_data_if_needed(
            self.resource['metadata'],
            'Response Headers',
            dict(response.items())
        )

        self.resource['metadata']['status_code'] = response.status_code

        if response.status_code >= 500:
            self.set_error()
